import pytest
from mau.errors import MauErrorException
from mau.lexers.main_lexer import MainLexer
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_arguments_empty():
    source = """
    []
    """

    assert runner(source).attributes_manager.attributes == ([], {}, [], None)


def test_arguments_subtype():
    source = """
    [*subtype]
    """

    assert runner(source).attributes_manager.attributes == ([], {}, [], "subtype")


def test_arguments_multiple_subtypes():
    source = """
    [*subtype1, *subtype2]
    """

    with pytest.raises(MauErrorException):
        assert runner(source).attributes_manager.attributes == ([], {}, [], "subtype1")


def test_arguments_args():
    source = """
    [arg1,arg2]
    """

    assert runner(source).attributes_manager.attributes == (
        ["arg1", "arg2"],
        {},
        [],
        None,
    )


def test_arguments_kwargs():
    source = """
    [key1=value1,key2=value2]
    """

    assert runner(source).attributes_manager.attributes == (
        [],
        {"key1": "value1", "key2": "value2"},
        [],
        None,
    )


def test_arguments_support_variables():
    source = """
    :arg1:number1
    :value1:42

    [{arg1},key1={value1}]
    """

    assert runner(source).attributes_manager.attributes == (
        ["number1"],
        {"key1": "42"},
        [],
        None,
    )


def test_arguments_tags():
    source = """
    [#tag1,#tag2]
    """

    assert runner(source).attributes_manager.attributes == (
        [],
        {},
        ["tag1", "tag2"],
        None,
    )



# pylint: disable=unused-import

from mau.environment.environment import Environment
from mau.lexers.main_lexer import MainLexer
from mau.lexers.text_lexer import TextLexer
from mau.parsers.main_parser import MainParser
from mau.text_buffer.context import Context
from mau.text_buffer.text_buffer import TextBuffer
from mau.visitors.base_visitor import BaseVisitor


# pylint: disable=import-outside-toplevel
def load_visitors():
    """
    This function loads all the visitors belonging to
    the group "mau.visitors". This code has been isolated
    in a function to allow visitor modules to import the
    Mau package without creating a cycle.
    """

    import sys

    if sys.version_info < (3, 10):
        from importlib_metadata import entry_points
    else:
        from importlib.metadata import entry_points

    discovered_plugins = entry_points(group="mau.visitors")

    # Load the available visitors
    visitors = [i.load() for i in discovered_plugins]
    visitors.append(BaseVisitor)

    return visitors


class ConfigurationError(ValueError):
    """Used to signal an error in the configuration"""


class Mau:
    def __init__(
        self,
        input_file_name,
        environment=None,
    ):
        # This is needed to set up the initial context for the lexer
        self.input_file_name = input_file_name

        # This will contain all the variables declared
        # in the text and in the configuration
        self.environment = environment or Environment()

        self.lexer = MainLexer(self.environment)
        self.parser = MainParser(self.environment)

    def run_lexer(self, text):
        context = Context(source=self.input_file_name)
        text_buffer = TextBuffer(text, context)

        self.lexer.process(text_buffer)

    def run_parser(self, tokens):
        self.parser.parse(tokens)
        self.parser.finalise()

    def create_visitor(self):
        visitor_class = self.environment.getvar("mau.visitor.class")

        return visitor_class(
            environment=self.environment,
        )

    def run_visitor(self, node):
        visitor = self.create_visitor()

        # Visit the document AST
        return visitor.visit(node)



from mau.lexers.text_lexer import TextLexer
from mau.nodes.inline import TextNode, VerbatimNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_verbatim():
    source = "`Many different words`"

    expected = [
        VerbatimNode("Many different words"),
    ]

    assert runner(source).nodes == expected


def test_verbatim_escaped_backtick():
    source = r"`\``"

    expected = [
        VerbatimNode("`"),
    ]

    assert runner(source).nodes == expected


def test_verbatim_style_inside_verbatim():
    source = r"`_Many different words_`"

    expected = [
        VerbatimNode("_Many different words_"),
    ]

    assert runner(source).nodes == expected


def test_verbatim_open():
    source = r"`Many different words"

    expected = [
        TextNode("`Many different words"),
    ]

    assert runner(source).nodes == expected


def test_verbatim_escape_characters():
    source = r"`$Many$ %different% \words`"

    expected = [
        VerbatimNode(r"$Many$ %different% \words"),
    ]

    assert runner(source).nodes == expected



from mau.environment.environment import Environment
from mau.nodes.header import HeaderNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.visitors.jinja_visitor import JinjaVisitor


def test_page_header_node():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "header.j2": (
            "{{ value }} - {{ level }} - {{ anchor }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = HeaderNode(
        value=SentenceNode(children=[TextNode("Just some text")]),
        level="3",
        anchor="someanchor",
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert (
        result
        == "Just some text - 3 - someanchor - arg1,arg2 - key1:value1 - tag1,tag2"
    )



from mau.lexers.main_lexer import MainLexer
from mau.nodes.block import BlockNode
from mau.nodes.inline import RawNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_block_raw_engine():
    source = """
    [*block, engine=raw]
    ----
    Raw content
    on multiple lines
    ----
    Secondary content
    on multiple lines as well
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype="block",
            children=[
                RawNode("Raw content"),
                RawNode("on multiple lines"),
            ],
            secondary_children=[
                RawNode("Secondary content"),
                RawNode("on multiple lines as well"),
            ],
            classes=[],
            title=None,
            engine="raw",
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]



from mau.nodes.page import ContainerNode, DocumentNode, HorizontalRuleNode


def test_horizontal_rule_node():
    node = HorizontalRuleNode(
        subtype="type1",
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.subtype == "type1"
    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "horizontal_rule"
    assert node == HorizontalRuleNode(
        subtype="type1",
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )


def test_container_node():
    node = ContainerNode()

    assert node.children == []
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "container"
    assert node == ContainerNode()


def test_document_node():
    node = DocumentNode()

    assert node.children == []
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "document"
    assert node == DocumentNode()



from mau.nodes.inline import TextNode
from mau.nodes.lists import ListItemNode, ListNode


def test_list_item_node():
    node = ListItemNode(
        level="3", children=[TextNode("This is a list with one element")]
    )

    assert node.level == "3"
    assert node.children == [TextNode("This is a list with one element")]
    assert node.node_type == "list_item"
    assert node == ListItemNode(
        level="3", children=[TextNode("This is a list with one element")]
    )


def test_list_node():
    node = ListNode(ordered=True, main_node=False)

    assert node.ordered is True
    assert node.children == []
    assert node.main_node is False
    assert node.start == 1
    assert node.subtype is None
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "list"
    assert node == ListNode(ordered=True, main_node=False)


def test_list_node_with_start():
    node = ListNode(ordered=True, main_node=False, start=42)

    assert node.ordered is True
    assert node.children == []
    assert node.main_node is False
    assert node.start == 42
    assert node.subtype is None
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "list"
    assert node == ListNode(ordered=True, main_node=False, start=42)



from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import TextNode
from mau.nodes.lists import ListItemNode, ListNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_parse_list_with_one_item():
    source = """
    * This is a list with one element
    """

    assert runner(source).nodes == [
        ListNode(
            ordered=False,
            main_node=True,
            children=[
                ListItemNode(
                    level="1", children=[TextNode("This is a list with one element")]
                )
            ],
        )
    ]


def test_parse_list_with_multiple_items():
    source = """
    * Item 1
    * Item 2
    """

    assert runner(source).nodes == [
        ListNode(
            ordered=False,
            main_node=True,
            children=[
                ListItemNode(level="1", children=[TextNode("Item 1")]),
                ListItemNode(level="1", children=[TextNode("Item 2")]),
            ],
        )
    ]


def test_parse_list_with_multiple_levels():
    source = """
    * Item 1
    ** Item 1.1
    * Item 2
    """

    assert runner(source).nodes == [
        ListNode(
            ordered=False,
            main_node=True,
            children=[
                ListItemNode(
                    level="1",
                    children=[
                        TextNode("Item 1"),
                        ListNode(
                            ordered=False,
                            main_node=False,
                            children=[
                                ListItemNode(
                                    level="2",
                                    children=[
                                        TextNode("Item 1.1"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ListItemNode(level="1", children=[TextNode("Item 2")]),
            ],
        )
    ]


def test_parse_numbered_list():
    source = """
    # Item 1
    ## Item 1.1
    # Item 2
    """

    assert runner(source).nodes == [
        ListNode(
            ordered=True,
            main_node=True,
            children=[
                ListItemNode(
                    level="1",
                    children=[
                        TextNode("Item 1"),
                        ListNode(
                            ordered=True,
                            main_node=False,
                            children=[
                                ListItemNode(
                                    level="2",
                                    children=[
                                        TextNode("Item 1.1"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 2")],
                ),
            ],
        )
    ]


def test_parse_mixed_list():
    source = """
    * Item 1
    ## Item 1.1
    * Item 2
    """

    assert runner(source).nodes == [
        ListNode(
            ordered=False,
            main_node=True,
            children=[
                ListItemNode(
                    level="1",
                    children=[
                        TextNode("Item 1"),
                        ListNode(
                            ordered=True,
                            main_node=False,
                            children=[
                                ListItemNode(
                                    level="2",
                                    children=[
                                        TextNode("Item 1.1"),
                                    ],
                                )
                            ],
                        ),
                    ],
                ),
                ListItemNode(level="1", children=[TextNode("Item 2")]),
            ],
        )
    ]


def test_parse_mixed_list_cannot_change_type():
    source1 = """
    * Item 1
    ## Item 1.1
    # Item 2
    """

    source2 = """
    * Item 1
    ## Item 1.1
    * Item 2
    """

    assert runner(source1).nodes == runner(source2).nodes


def test_parse_numbered_list_continue():
    source = """
    # Item 1
    # Item 2

    [start=auto]
    # Item 3
    # Item 4
    """

    assert runner(source).nodes == [
        ListNode(
            ordered=True,
            main_node=True,
            start=1,
            children=[
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 1")],
                ),
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 2")],
                ),
            ],
        ),
        ListNode(
            ordered=True,
            main_node=True,
            start=3,
            children=[
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 3")],
                ),
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 4")],
                ),
            ],
            kwargs={},
        ),
    ]


def test_parse_numbered_list_continue_after_forced():
    source = """
    # Item 1
    # Item 2

    [start=20]
    # Item 20
    # Item 21

    [start=auto]
    # Item 22
    # Item 23
    """

    assert runner(source).nodes == [
        ListNode(
            ordered=True,
            main_node=True,
            start=1,
            children=[
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 1")],
                ),
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 2")],
                ),
            ],
        ),
        ListNode(
            ordered=True,
            main_node=True,
            start=20,
            children=[
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 20")],
                ),
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 21")],
                ),
            ],
            kwargs={},
        ),
        ListNode(
            ordered=True,
            main_node=True,
            start=22,
            children=[
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 22")],
                ),
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 23")],
                ),
            ],
            kwargs={},
        ),
    ]


def test_parse_numbered_list_do_not_continue():
    source = """
    # Item 1
    # Item 2

    # Item 3
    # Item 4
    """

    assert runner(source).nodes == [
        ListNode(
            ordered=True,
            main_node=True,
            children=[
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 1")],
                ),
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 2")],
                ),
            ],
        ),
        ListNode(
            ordered=True,
            main_node=True,
            children=[
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 3")],
                ),
                ListItemNode(
                    level="1",
                    children=[TextNode("Item 4")],
                ),
            ],
        ),
    ]


def test_parse_list_with_subtype():
    source = """
    [*type1]
    * This is a list with one element
    """

    assert runner(source).nodes == [
        ListNode(
            ordered=False,
            main_node=True,
            children=[
                ListItemNode(
                    level="1", children=[TextNode("This is a list with one element")]
                )
            ],
            subtype="type1",
        )
    ]



import functools
import logging

import yaml
from mau.errors import MauError, MauErrorException

_logger = logging.getLogger(__name__)


class MauVisitorError(MauError):
    source = "visitor"

    def print_details(self):  # pragma: no cover
        super().print_details()

        print("Node:")
        print(self.details["node"])


class NoAliasDumper(yaml.SafeDumper):  # pragma: no cover
    # pylint: disable=unused-argument
    def ignore_aliases(self, data):
        return True


no_aliases_dump = functools.partial(yaml.dump, Dumper=NoAliasDumper)


class BaseVisitor:
    format_code = "yaml"
    extension = ""
    transform = no_aliases_dump

    _join_with = {}
    _join_with_default = None

    # pylint: disable=unused-argument
    def __init__(self, environment, *args, **kwds):
        self.toc = None
        self.footnotes = None

        self.environment = environment

    def visit(self, node, *args, **kwargs):
        # The visitor has to define functions for each node type.
        # Those shall return a dictionary with the key "data" that
        # contains the result of the visit.
        # This is done to provide space for metadata or other values
        # like templates used to render the node.

        if node is None:
            return {}

        return node.accept(self, *args, **kwargs)

    def visitlist(self, node, nodes, *args, **kwargs):
        join_with = self._join_with.get(node.node_type, self._join_with_default)

        visited_nodes = [self.visit(i, *args, **kwargs) for i in nodes]

        if join_with is not None:
            return join_with.join(visited_nodes)

        return visited_nodes

    def visitdict(self, node, nodedict, *args, **kwargs):
        return {k: self.visit(v, *args, **kwargs) for k, v in nodedict.items()}

    def _error(self, message, node=None):
        error = MauVisitorError(
            message=message,
            details={
                "node": node,
            },
        )

        raise MauErrorException(error)

    def _visit_default(self, node, *args, **kwargs):
        return {
            "data": {
                "type": node.node_type,
                "subtype": node.subtype,
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    # Inline nodes

    def _visit_raw(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "value": node.value,
            }
        )

        return result

    def _visit_text(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "value": node.value,
            }
        )

        return result

    def _visit_verbatim(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "value": node.value,
            }
        )

        return result

    def _visit_sentence(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "content": self.visitlist(node, node.children, *args, **kwargs),
            }
        )

        return result

    def _visit_style(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "value": node.value,
                "content": self.visitlist(node, node.children, *args, **kwargs),
            }
        )

        return result

    def _visit_macro(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "name": node.name,
            }
        )

        return result

    def _visit_footnote(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "content": self.visitlist(node, node.children, *args, **kwargs),
                "number": node.number,
                "reference_anchor": node.reference_anchor,
                "content_anchor": node.content_anchor,
            }
        )

        return result

    def _visit_macro__class(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "classes": node.classes,
                "content": self.visitlist(node, node.children, *args, **kwargs),
            }
        )

        return result

    def _visit_macro__link(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "target": node.target,
                "content": self.visitlist(node, node.children, *args, **kwargs),
            }
        )

        return result

    def _visit_macro__header(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "header": {
                    "anchor": node.header.anchor,
                    "value": self.visit(node.header.value, *args, **kwargs),
                    "level": node.header.level,
                },
                "content": self.visitlist(node, node.children, *args, **kwargs)
                if node.children
                else self.visit(node.header.value, *args, **kwargs),
            }
        )

        return result

    def _visit_macro__image(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "uri": node.uri,
                "alt_text": node.alt_text,
                "width": node.width,
                "height": node.height,
            }
        )

        return result

    def _visit_list_item(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "level": int(node.level),
                "content": self.visitlist(node, node.children, *args, **kwargs),
            }
        )

        return result

    def _visit_paragraph(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "title": self.visit(node.title, *args, **kwargs),
                "content": self.visitlist(node, node.children, *args, **kwargs),
            }
        )

        return result

    def _visit_header(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)

        result["data"].update(
            {
                "value": self.visit(node.value, *args, **kwargs),
                "level": int(node.level),
                "anchor": node.anchor,
            }
        )

        return result

    def _visit_list(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "ordered": node.ordered,
                "items": self.visitlist(node, node.children, *args, **kwargs),
                "main_node": node.main_node,
                "start": node.start,
            }
        )

        return result

    def _visit_content(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "content_type": node.content_type,
                "uris": node.uris,
                "title": self.visit(node.title, *args, **kwargs),
            }
        )

        return result

    def _visit_content_image(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "uri": node.uri,
                "alt_text": node.alt_text,
                "classes": node.classes,
                "title": self.visit(node.title, *args, **kwargs),
            }
        )

        return result

    def _visit_block(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "content": self.visitlist(node, node.children, *args, **kwargs),
                "secondary_content": self.visitlist(
                    node, node.secondary_children, *args, **kwargs
                ),
                "subtype": node.subtype,
                "classes": node.classes,
                "title": self.visit(node.title, *args, **kwargs),
                "engine": node.engine,
                "preprocessor": node.preprocessor,
            }
        )

        return result

    def _visit_source(self, node, *args, **kwargs):
        for method_name in [
            "_visit_source__default",
            f"_visit_source_{node.language}",
        ]:
            try:
                method = getattr(self, method_name)
            except AttributeError:
                continue

        return method(node)

    def _visit_source__default(self, node, *args, **kwargs):
        code = self.visitlist(node, node.code, *args, **kwargs)

        # This is a list of callouts for each line
        # None if the callout is not present, otherwise
        # the rendered callout
        markers = [None] * len(code)

        for marker in node.markers:
            markers[marker.line] = self.visit(marker, *args, **kwargs)

        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "code": code,
                "subtype": node.subtype,
                "language": node.language,
                "callouts": self.visitlist(node, node.callouts, *args, **kwargs),
                "highlights": node.highlights,
                "markers": markers,
                "classes": node.classes,
                "title": self.visit(node.title, *args, **kwargs),
                "preprocessor": node.preprocessor,
                "lines": len(node.code),
            }
        )

        return result

    def _visit_callout(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "line": node.line,
                "marker": node.marker,
            }
        )

        return result

    def _visit_callouts_entry(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "marker": node.marker,
                "value": node.value,
            }
        )

        return result

    def _visit_toc_entry(self, node, *args, **kwargs):
        return {
            "data": {
                "type": node.node_type,
                "value": self.visit(node.value, *args, **kwargs),
                "anchor": node.anchor,
                "children": self.visitlist(node, node.children, *args, **kwargs),
                "tags": node.tags,
            },
        }

    def _visit_toc(self, node, *args, **kwargs):
        return {
            "data": {
                "type": node.node_type,
                "entries": self.visitlist(node, node.children, *args, **kwargs),
                "args": node.args,
                "kwargs": node.kwargs,
                "tags": node.tags,
            },
        }

    def _visit_footnotes(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "entries": self.visitlist(node, node.children, *args, **kwargs),
            }
        )

        return result

    def _visit_footnotes_entry(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "content": self.visitlist(node, node.children, *args, **kwargs),
                "number": node.number,
                "reference_anchor": node.reference_anchor,
                "content_anchor": node.content_anchor,
            }
        )

        return result

    def _visit_container(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "content": self.visitlist(node, node.children, *args, **kwargs),
            }
        )

        return result

    def _visit_document(self, node, *args, **kwargs):
        return self._visit_container(node)

    def _visit_block_group(self, node, *args, **kwargs):
        result = self._visit_default(node, *args, **kwargs)
        result["data"].update(
            {
                "title": self.visit(node.title, *args, **kwargs),
                "group_name": node.group_name,
                "group": self.visitdict(node, node.group, *args, **kwargs),
            }
        )

        return result



class InternalLinksManager:
    def __init__(self, parser):
        # This list containes the internal links created
        # in the text through a macro.
        self.links = []

        # This dictionary contains the headers
        # flagged with an id
        self.headers = {}

        # This is the parser that contains the manager
        self.parser = parser

    def add_header(self, header_id, node):
        if header_id in self.headers:
            self.parser._error(f"Duplicate header id detected: {header_id}")

        self.headers[header_id] = node

    def process_links(self):
        for link in self.links:
            try:
                link.header = self.headers[link.header_id]
            except KeyError:
                self.parser._error(f"Cannot find header with id {link.header_id}")

    def update(self, other):
        self.update_links(other.links)
        self.headers.update(other.headers)

    def update_links(self, links):
        self.links.extend(links)



import pytest
import pathlib


import textwrap

from mau.environment.environment import Environment
from mau.text_buffer.context import Context
from mau.text_buffer.text_buffer import TextBuffer


def dedent(text):
    return textwrap.dedent(text).strip()


def init_lexer_factory(lexer_class):
    """
    A factory that returns a lexer initialiser.
    The returned function initialises the lexer
    and returns it.
    """

    def _init_lexer(environment=None):
        return lexer_class(environment)

    return _init_lexer


def lexer_runner_factory(lexer_class, *args, **kwds):
    """
    A factory that returns a lexer runner.
    The returned function initialises and
    runs the lexer on the given source.
    """

    init_lexer = init_lexer_factory(lexer_class)

    def _run(source, environment=None, **kwargs):
        kwds.update(kwargs)

        environment = environment or Environment()

        text_buffer = TextBuffer(textwrap.dedent(source), Context())

        lexer = init_lexer(environment, *args, **kwds)
        lexer.process(text_buffer)

        return lexer

    return _run


def init_parser_factory(lexer_class, parser_class):
    """
    A factory that returns a parser initialiser.
    The returned function initialises and runs the lexer,
    initialises the parser and returns it.
    """

    def _init_parser(text, environment=None, *args, **kwargs):
        text_buffer = TextBuffer(text, Context(source="main"))

        lex = lexer_class(environment)
        lex.process(text_buffer)

        par = parser_class(environment, *args, **kwargs)
        par.tokens = lex.tokens

        return par

    return _init_parser


def parser_runner_factory(lexer_class, parser_class, *args, **kwds):
    """
    A factory that returns a parser runner.
    The returned function runs the parser on the given source.
    """

    init_parser = init_parser_factory(lexer_class, parser_class)

    def _run(source, environment=None, **kwargs):
        kwds.update(kwargs)

        environment = environment or Environment()

        parser = init_parser(textwrap.dedent(source), environment, *args, **kwds)
        parser.parse(parser.tokens)
        parser.finalise()

        return parser

    return _run


def load_files(source_dir, source_ext, expected_dir, expected_ext):
    source_files = pathlib.Path(source_dir).rglob(f"*.{source_ext}")

    files = [
        (i, expected_dir / i.with_suffix(f".{expected_ext}").relative_to(source_dir))
        for i in source_files
    ]

    return files


def collect_test_files(
    tests_dir, source_dir_name, source_ext, expected_dir_name, expected_ext
):
    source_dir = tests_dir / source_dir_name
    expected_dir = tests_dir / expected_dir_name

    return [
        pytest.param(i, j, id=str(i.relative_to(tests_dir)))
        for i, j in load_files(source_dir, source_ext, expected_dir, expected_ext)
    ]



from unittest.mock import Mock

from mau.nodes.nodes import Node, ValueNode


def test_node():
    node = Node()

    assert node.node_type == "node"
    assert node.asdict() == {
        "type": "node",
        "subtype": None,
        "children": [],
        "args": [],
        "kwargs": {},
        "tags": [],
    }


def test_value_node():
    node = ValueNode("somevalue")

    assert node.value == "somevalue"
    assert node.node_type == "value_node"
    assert node == ValueNode("somevalue")


def test_node_accept():
    visitor = Mock()
    node = Node()
    node.node_type = "mynode"

    assert node.accept(visitor) == visitor._visit_mynode()


def test_node_accept_default():
    # This makes the Mock raise an AttributeError
    # for anything that is not in the list
    visitor = Mock(spec=["_visit_default"])
    node = Node()
    node.node_type = "mynode"

    assert node.accept(visitor) == visitor._visit_default()



from mau.nodes.nodes import Node


class BlockNode(Node):
    """A block.

    This node contains a generic block.

    Arguments:
        blocktype: the type of this block
        content: content of the block
        secondary_content: secondary content of this block
        title: title of this block
        classes: a comma-separated list of classes
        engine: the engine used to render this block
        preprocessor: the preprocessor used for this block
        args: unnamed arguments
        kwargs: named arguments
        tags: tags
    """

    node_type = "block"

    def __init__(
        self,
        classes=None,
        title=None,
        engine=None,
        preprocessor=None,
        parent=None,
        parent_position=None,
        children=None,
        secondary_children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.classes = classes or []
        self.title = title
        self.engine = engine
        self.preprocessor = preprocessor
        self.secondary_children = secondary_children or []

    def _custom_dict(self):
        return {
            "classes": self.classes,
            "title": self.title,
            "engine": self.engine,
            "preprocessor": self.preprocessor,
            "secondary_children": self.secondary_children,
        }


class BlockGroupNode(Node):
    """This instructs Mau to insert a group of nodes."""

    node_type = "block_group"

    def __init__(
        self,
        group_name,
        group,
        title=None,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.title = title
        self.group_name = group_name
        self.group = group

    def _custom_dict(self):
        return {
            "title": self.title,
            "group_name": self.group_name,
            "group": self.group,
        }



from mau.environment.environment import Environment
from mau.nodes.inline import TextNode
from mau.nodes.lists import ListItemNode, ListNode
from mau.visitors.jinja_visitor import JinjaVisitor


def test_inline_list_item_node():
    templates = {
        "text.j2": "{{ value }}",
        "list_item.j2": "{{ level }} - {{ content }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = ListItemNode(level="4", children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == "4 - Just some text."


def test_page_list_node():
    templates = {
        "text.j2": "{{ value }}",
        "list_item.j2": "{{ level }}:{{ content }}",
        "list.j2": (
            "{{ ordered }} - {{ items }} - {{ main_node }} - {{ args | join(',') }} - "
            "{{ kwargs|items|map('join', '=')|join(',') }} - {{ tags | join(',') }} - "
            "{{ kwargs.start }}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1", "start": 4}
    tags = ["tag1", "tag2"]
    node = ListNode(
        ordered=True,
        main_node=True,
        children=[ListItemNode(level="4", children=[TextNode("Just some text.")])],
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert (
        result
        == "True - 4:Just some text. - True - arg1,arg2 - key1=value1,start=4 - tag1,tag2 - 4"
    )



import logging
from pathlib import Path

import jinja2
from mau.environment.environment import Environment

# from mau.parsers.toc import create_toc
from mau.visitors.base_visitor import BaseVisitor

_logger = logging.getLogger(__name__)

JOIN_CHARACTERS = {
    "block": "\n",
    "footnote": "\n",
    "footnotes_entry": "\n",
    "source": None,
}


# pylint: disable=import-outside-toplevel
def load_template_providers():  # pragma: no cover
    """
    This function loads all the template providers belonging to
    the group "mau.templates".
    """

    import sys

    if sys.version_info < (3, 10):
        from importlib_metadata import entry_points
    else:
        from importlib.metadata import entry_points

    discovered_plugins = entry_points(group="mau.templates")

    # Load the available plugins
    return {i.name: i.load() for i in discovered_plugins}


def load_templates_from_path(path, filt=None):  # pragma: no cover
    """
    Loads templates fromt he given path into the dictionary `output`.
    The path is expected to contain files named after the template
    they contain and subdirectories with the same structure.
    """

    result = {}

    for obj in path.iterdir():
        if obj.is_file():
            result[obj.name] = filt(obj.read_text()) if filt else obj.read_text()
        else:
            result[obj.name] = load_templates_from_path(obj, filt=filt)

    return result


def create_templates(prefixes, node_templates, node, extension=None):
    prefixes = [f"{prefix}." for prefix in prefixes]
    prefixes.append("")

    node_templates.append(node.node_type)

    # Build
    # [
    # "parent_type.parent_subtype.position",
    # "parent_type.parent_subtype.",
    # "parent_type.position",
    # "parent_type.",
    # ""
    # ]

    parent_types = [""]
    if node.parent:
        parent_types.append(f"{node.parent.node_type}.")

        if node.parent_position:
            parent_types.append(f"{node.parent.node_type}.{node.parent_position}.")

        if node.parent.subtype:
            parent_types.append(f"{node.parent.node_type}.{node.parent.subtype}.")

            if node.parent_position:
                parent_types.append(
                    f"{node.parent.node_type}.{node.parent.subtype}.{node.parent_position}."
                )
    parent_types = parent_types[::-1]

    # Build [".{node_subtype}", ""]
    node_subtypes = [""]
    if node.subtype:
        node_subtypes.append(f".{node.subtype}")
    node_subtypes = node_subtypes[::-1]

    # Build [".{tag1}", ".{tag2}", ..., ""]
    node_tags = [f".{tag}" for tag in node.tags]
    node_tags.append("")

    templates = [
        f"{prefix}{parent_type}{node_template}{node_subtype}{node_tag}"
        for prefix in prefixes
        for parent_type in parent_types
        for node_template in node_templates
        for node_subtype in node_subtypes
        for node_tag in node_tags
    ]

    # The template full name contains the extension
    # with the type of template, e.g. document.txt
    if extension:
        templates = [f"{template}.{extension}" for template in templates]

    return templates


class TemplateNotFound(ValueError):
    pass


class JinjaVisitor(BaseVisitor):
    format_code = "jinja"
    extension = "j2"
    transform = None
    templates_filter = None

    environment_options = {}
    default_templates = Environment()

    _join_with = JOIN_CHARACTERS
    _join_with_default = ""

    def __init__(
        self,
        environment,
        *args,
        **kwds,
    ):
        super().__init__(environment)

        self.template_prefixes = self.environment.getvar("mau.visitor.prefixes", [])

        # Load templates from the visitor plugin
        self.templates = Environment(self.default_templates)

        # Load templates from template provider plugins
        available_providers = load_template_providers()
        requested_providers = environment.getvar("mau.visitor.template_providers", [])
        for provider in requested_providers:  # pragma: no cover
            if provider not in available_providers:
                print(f"Template provider {provider} is not available")
                continue

            self.templates.update(available_providers[provider].templates)

        # Load custom templates provided as files
        templates_directory = environment.getvar("mau.visitor.templates_directory")
        if templates_directory:  # pragma: no cover
            self.templates.update(
                load_templates_from_path(
                    Path.cwd() / Path(templates_directory),
                    filt=self.templates_filter,
                )
            )

        # Load custom templates provided as a dictionary
        self.templates.update(
            environment.getvar("mau.visitor.custom_templates", Environment())
        )

        # These act as a temporary storage while we are in mau blocks.
        # There, ToC and footnotes have to be isolated, so at the
        # beginning of the block we save the main ToC and footnotes
        self._temp_toc = None
        self._temp_footnotes = None

        # This is the environment that uses the internal dict
        self._dict_env = jinja2.Environment(
            loader=jinja2.DictLoader(self.templates.asflatdict()),
            **self.environment_options,
        )

    def _render(self, template_full_name, **kwargs):
        # This renders a template using the current
        # environment and the given parameters
        try:
            template = self._dict_env.get_template(template_full_name)
        except jinja2.exceptions.TemplateNotFound as exception:
            raise TemplateNotFound(exception) from exception

        return template.render(config=self.environment.asdict(), **kwargs)

    def visit(self, node, *args, **kwargs):
        # The visitor has to define functions for each node type
        # and those shall return a dictionary of keys.
        #
        # The node types are a list made of the key `node_types` and the node type.
        # This allows a function to return one or more types that have to
        # be used instead of the standard type, which also allows to write
        # generic functions. See _visit_style in HTMLVisitor which works for
        # multiple styles.
        #
        # The rest of the returned keys are passed to the _render function
        # as keys and are thus available in the template.

        # Template names are created with this schema

        # [prefix.][parent_type.][parent_subtype.][parent_position.][node_template][.node_subtype][.ext]

        if node is None:
            return {}

        result = super().visit(node, *args, **kwargs)
        node_templates = result.get("templates", [])

        templates = create_templates(
            self.template_prefixes, node_templates, node, self.extension
        )

        # The key "data" contains the values that we want
        # to pass to the template. These are used as arguments
        # when calling the template.
        data = result["data"]

        for template in templates:
            try:
                return self._render(template, **data)
            except TemplateNotFound:
                continue

        self._error(f"Cannot find any of the following templates {templates}", node)

        return None  # pragma: no cover

    def _visit_style(self, node, *args, **kwargs):
        base = super()._visit_style(node)
        base["templates"] = [f"style.{node.value}"]

        return base

    def _visit_macro(self, node, *args, **kwargs):
        base = super()._visit_macro(node)
        base["templates"] = [f"macro.{node.name}"]

        return base

    def _visit_content(self, node, *args, **kwargs):
        base = super()._visit_content(node)
        base["templates"] = [
            f"content.{node.content_type}",
        ]

        return base

    def _visit_block(self, node, *args, **kwargs):
        base = super()._visit_block(node)

        base["templates"] = []
        if node.engine is not None:
            base["templates"].append(f"{node.node_type}.{node.engine}")

        return base

    def _visit_block_group(self, node, *args, **kwargs):
        base = super()._visit_block_group(node)

        base["templates"] = [
            f"block_group.{node.group_name}",
        ]

        return base

    def _visit_source(self, node, *args, **kwargs):
        base = super()._visit_source(node)
        base["templates"] = [
            f"source.{node.language}",
        ]

        return base



from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.macros import MacroLinkNode
from mau.nodes.paragraph import ParagraphNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_parse_paragraphs():
    source = """
    This is a paragraph.
    This is part of the same paragraph.

    This is a new paragraph.
    """

    assert runner(source).nodes == [
        ParagraphNode(
            children=[
                TextNode("This is a paragraph. This is part of the same paragraph."),
            ],
            args=[],
            kwargs={},
        ),
        ParagraphNode(
            children=[
                TextNode("This is a new paragraph."),
            ],
            args=[],
            kwargs={},
        ),
    ]


def test_parse_paragraph_starting_with_a_macro():
    source = "[link](http://some.where,This) is the link I want"

    assert runner(source).nodes == [
        ParagraphNode(
            children=[
                MacroLinkNode(target="http://some.where", children=[TextNode("This")]),
                TextNode(" is the link I want"),
            ]
        )
    ]


def test_attributes_paragraph():
    source = """
    [*type, arg1,key1=value1]
    This is text
    """

    assert runner(source).nodes == [
        ParagraphNode(
            children=[TextNode("This is text")],
            args=["arg1"],
            kwargs={"key1": "value1"},
            subtype="type",
        ),
    ]


def test_paragraph_title():
    source = """
    .Title
    This is text
    """

    parser = runner(source)

    assert parser.nodes == [
        ParagraphNode(
            children=[TextNode("This is text")],
            title=SentenceNode(children=[TextNode("Title")]),
        ),
    ]

    paragraph_node = parser.nodes[0]
    text_node = paragraph_node.title.children[0]

    assert text_node.parent == paragraph_node
    assert text_node.parent_position == "title"



from mau.nodes.header import HeaderNode
from mau.nodes.toc import TocEntryNode
from mau.parsers.toc import create_toc


def test_create_toc():
    headers_list = [
        HeaderNode("Header 1", "1", "header-1"),
        HeaderNode("Header 1.1", "2", "header-1-1"),
        HeaderNode("Header 1.2", "2", "header-1-2"),
        HeaderNode("Header 2", "1", "header-2"),
        HeaderNode("Header 2.1", "2", "header-2-1"),
        HeaderNode("Header 2.1.1", "3", "header-2-1-1"),
    ]

    assert create_toc(headers_list) == [
        TocEntryNode(
            value="Header 1",
            anchor="header-1",
            children=[
                TocEntryNode(
                    value="Header 1.1",
                    anchor="header-1-1",
                    children=[],
                ),
                TocEntryNode(
                    value="Header 1.2",
                    anchor="header-1-2",
                    children=[],
                ),
            ],
        ),
        TocEntryNode(
            value="Header 2",
            anchor="header-2",
            children=[
                TocEntryNode(
                    value="Header 2.1",
                    anchor="header-2-1",
                    children=[
                        TocEntryNode(
                            value="Header 2.1.1",
                            anchor="header-2-1-1",
                            children=[],
                        ),
                    ],
                ),
            ],
        ),
    ]


def test_create_toc_exclude_top_level():
    headers_list = [
        HeaderNode("Header 1", "1", "header-1", tags=["notoc"]),
        HeaderNode("Header 1.1", "2", "header-1-1"),
        HeaderNode("Header 1.2", "2", "header-1-2"),
        HeaderNode("Header 2", "1", "header-2"),
        HeaderNode("Header 2.1", "2", "header-2-1"),
        HeaderNode("Header 2.1.1", "3", "header-2-1-1"),
    ]

    assert create_toc(headers_list, exclude_tag="notoc") == [
        TocEntryNode(
            value="Header 2",
            anchor="header-2",
            children=[
                TocEntryNode(
                    value="Header 2.1",
                    anchor="header-2-1",
                    children=[
                        TocEntryNode(
                            value="Header 2.1.1",
                            anchor="header-2-1-1",
                            children=[],
                        ),
                    ],
                ),
            ],
        ),
    ]


def test_create_toc_exclude_lower_level():
    headers_list = [
        HeaderNode("Header 1", "1", "header-1"),
        HeaderNode("Header 1.1", "2", "header-1-1"),
        HeaderNode("Header 1.2", "2", "header-1-2"),
        HeaderNode("Header 2", "1", "header-2"),
        HeaderNode("Header 2.1", "2", "header-2-1", tags=["notoc"]),
        HeaderNode("Header 2.1.1", "3", "header-2-1-1"),
    ]

    assert create_toc(headers_list, exclude_tag="notoc") == [
        TocEntryNode(
            value="Header 1",
            anchor="header-1",
            children=[
                TocEntryNode(
                    value="Header 1.1",
                    anchor="header-1-1",
                    children=[],
                ),
                TocEntryNode(
                    value="Header 1.2",
                    anchor="header-1-2",
                    children=[],
                ),
            ],
        ),
        TocEntryNode(
            value="Header 2",
            anchor="header-2",
            children=[],
        ),
    ]


def test_create_toc_orphan_nodes():
    headers_list = [
        HeaderNode("Header 1", "1", "header-1"),
        HeaderNode("Header 1.1.1", "3", "header-1-1-1"),
        HeaderNode("Header 1.2", "2", "header-1-2"),
    ]

    assert create_toc(headers_list) == [
        TocEntryNode(
            value="Header 1",
            anchor="header-1",
            children=[
                TocEntryNode(
                    value="Header 1.1.1",
                    anchor="header-1-1-1",
                    children=[],
                ),
                TocEntryNode(
                    value="Header 1.2",
                    anchor="header-1-2",
                    children=[],
                ),
            ],
        ),
    ]



from mau.nodes.nodes import Node, ValueNode


class UnnamedArgumentNode(ValueNode):
    """
    This node contains an unnamed argument.
    """

    node_type = "unnamed_argument"


class NamedArgumentNode(Node):
    """
    This node contains a named argument.
    """

    node_type = "named_argument"

    def __init__(
        self,
        key,
        value,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.key = key
        self.value = value

    def _custom_dict(self):
        return {
            "key": self.key,
            "value": self.value,
        }



from mau.nodes.inline import TextNode
from mau.nodes.paragraph import ParagraphNode


def test_paragraph_node():
    node = ParagraphNode(
        children=[TextNode("sometext")],
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.children == [TextNode("sometext")]
    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "paragraph"
    assert node == ParagraphNode(
        children=[TextNode("sometext")],
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )



from mau.environment.environment import Environment
from mau.nodes.block import BlockGroupNode, BlockNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.paragraph import ParagraphNode
from mau.visitors.base_visitor import BaseVisitor


def test_block_node_standard_block_template():
    visitor = BaseVisitor(Environment())

    node = BlockNode(
        subtype="someblock",
        children=[
            ParagraphNode(
                children=[
                    TextNode("my content"),
                ]
            ),
        ],
        secondary_children=[
            ParagraphNode(
                children=[
                    TextNode("my secondary content"),
                ]
            ),
        ],
        classes=["class1", "class2"],
        title=SentenceNode(children=[TextNode("sometitle")]),
        engine="someengine",
        preprocessor="somepreprocessor",
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "block",
            "subtype": "someblock",
            "content": [
                {
                    "data": {
                        "type": "paragraph",
                        "content": [
                            {
                                "data": {
                                    "type": "text",
                                    "value": "my content",
                                    "args": [],
                                    "kwargs": {},
                                    "subtype": None,
                                    "tags": [],
                                }
                            }
                        ],
                        "title": {},
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                }
            ],
            "secondary_content": [
                {
                    "data": {
                        "type": "paragraph",
                        "content": [
                            {
                                "data": {
                                    "type": "text",
                                    "value": "my secondary content",
                                    "args": [],
                                    "kwargs": {},
                                    "subtype": None,
                                    "tags": [],
                                }
                            }
                        ],
                        "title": {},
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                }
            ],
            "classes": ["class1", "class2"],
            "title": {
                "data": {
                    "content": [
                        {
                            "data": {
                                "type": "text",
                                "value": "sometitle",
                                "args": [],
                                "kwargs": {},
                                "subtype": None,
                                "tags": [],
                            }
                        },
                    ],
                    "type": "sentence",
                    "args": [],
                    "kwargs": {},
                    "subtype": None,
                    "tags": [],
                },
            },
            "engine": "someengine",
            "preprocessor": "somepreprocessor",
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }


def test_block_group():
    visitor = BaseVisitor(Environment())

    node = BlockGroupNode(
        group_name="somegroup",
        group={
            "left": BlockNode(
                subtype="sometype1",
                children=[
                    ParagraphNode(
                        children=[
                            TextNode("Block 1"),
                        ]
                    ),
                ],
                secondary_children=[],
                classes=[],
                title=None,
                engine="group",
                preprocessor="none",
                args=[],
                kwargs={},
            ),
            "right": BlockNode(
                subtype="sometype2",
                children=[
                    ParagraphNode(
                        children=[
                            TextNode("Block 2"),
                        ]
                    ),
                ],
                secondary_children=[],
                classes=[],
                title=None,
                engine="group",
                preprocessor="none",
                args=[],
                kwargs={},
            ),
        },
        args=[],
        kwargs={},
        tags=[],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "block_group",
            "group_name": "somegroup",
            "group": {
                "left": {
                    "data": {
                        "type": "block",
                        "subtype": "sometype1",
                        "content": [
                            {
                                "data": {
                                    "type": "paragraph",
                                    "content": [
                                        {
                                            "data": {
                                                "type": "text",
                                                "value": "Block 1",
                                                "args": [],
                                                "kwargs": {},
                                                "subtype": None,
                                                "tags": [],
                                            }
                                        }
                                    ],
                                    "title": {},
                                    "args": [],
                                    "kwargs": {},
                                    "subtype": None,
                                    "tags": [],
                                }
                            }
                        ],
                        "secondary_content": [],
                        "classes": [],
                        "title": {},
                        "engine": "group",
                        "preprocessor": "none",
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    },
                },
                "right": {
                    "data": {
                        "type": "block",
                        "subtype": "sometype2",
                        "content": [
                            {
                                "data": {
                                    "type": "paragraph",
                                    "content": [
                                        {
                                            "data": {
                                                "type": "text",
                                                "value": "Block 2",
                                                "args": [],
                                                "kwargs": {},
                                                "subtype": None,
                                                "tags": [],
                                            }
                                        }
                                    ],
                                    "title": {},
                                    "args": [],
                                    "kwargs": {},
                                    "subtype": None,
                                    "tags": [],
                                }
                            }
                        ],
                        "secondary_content": [],
                        "classes": [],
                        "title": {},
                        "engine": "group",
                        "preprocessor": "none",
                        "args": [],
                        "kwargs": {},
                        "tags": [],
                    },
                },
            },
            "title": {},
            "args": [],
            "kwargs": {},
            "tags": [],
            "subtype": None,
        }
    }



from mau.lexers.text_lexer import TextLexer
from mau.nodes.inline import StyleNode, TextNode
from mau.nodes.macros import MacroLinkNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_macro_link():
    source = '[link](https://somedomain.org/the/path, "link text")'

    expected = [
        MacroLinkNode(
            "https://somedomain.org/the/path", children=[TextNode("link text")]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_link_without_text():
    source = '[link]("https://somedomain.org/the/path")'

    expected = [
        MacroLinkNode(
            "https://somedomain.org/the/path",
            children=[TextNode("https://somedomain.org/the/path")],
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_link_with_rich_text():
    source = (
        '[link]("https://somedomain.org/the/path", "Some text with _styled words_")'
    )

    expected = [
        MacroLinkNode(
            "https://somedomain.org/the/path",
            children=[
                TextNode("Some text with "),
                StyleNode(value="underscore", children=[TextNode("styled words")]),
            ],
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_mailto():
    source = "[mailto](info@projectmau.org)"

    expected = [
        MacroLinkNode(
            "mailto:info@projectmau.org", children=[TextNode("info@projectmau.org")]
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_mailto_custom_text():
    source = '[mailto](info@projectmau.org, "my email")'

    expected = [
        MacroLinkNode("mailto:info@projectmau.org", children=[TextNode("my email")]),
    ]

    assert runner(source).nodes == expected



# Page nodes can be found at top level in a page

from mau.nodes.nodes import Node


class HorizontalRuleNode(Node):
    """A horizontal rule."""

    node_type = "horizontal_rule"


class ContainerNode(Node):
    node_type = "container"

    def clone(self):
        return self.__class__(
            parent=self.parent,
            parent_position=self.parent_position,
            children=self.children,
            subtype=self.subtype,
            args=self.args,
            kwargs=self.kwargs,
            tags=self.tags,
        )


class DocumentNode(ContainerNode):
    """A document.

    This node represents the full document.

    Arguments:
        content: the content of the document
    """

    node_type = "document"



from mau.environment.environment import Environment
from mau.nodes.content import ContentImageNode, ContentNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.visitors.jinja_visitor import JinjaVisitor


def test_page_content_node():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "content.sometype.j2": (
            "{{ title }} - "
            "{{ uris | join(',') }} - "
            "{{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = ContentNode(
        "sometype",
        uris=["/uri1", "/uri2"],
        title=SentenceNode(children=[TextNode("sometitle")]),
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == "sometitle - /uri1,/uri2 - arg1,arg2 - key1:value1 - tag1,tag2"


def test_content_image_node():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "content_image.j2": (
            "{{ uri }} - {{ alt_text }} - {{ classes | join(',') }} - "
            "{{ title }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = ContentImageNode(
        "someuri",
        "alttext",
        ["class1", "class2"],
        title=SentenceNode(children=[TextNode("sometitle")]),
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert (
        result
        == "someuri - alttext - class1,class2 - sometitle - arg1,arg2 - key1:value1 - tag1,tag2"
    )



from mau.environment.environment import Environment
from mau.nodes.block import BlockNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.paragraph import ParagraphNode
from mau.visitors.jinja_visitor import JinjaVisitor


def test_page_paragraph_node():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "paragraph.j2": (
            "{{ content }} - {{ title }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = ParagraphNode(
        title=SentenceNode(children=[TextNode("sometitle")]),
        children=[TextNode("Just some text")],
        args=args,
        kwargs=kwargs,
        tags=tags,
    )

    result = visitor.visit(node)

    assert result == "Just some text - sometitle - arg1,arg2 - key1:value1 - tag1,tag2"


def test_page_paragraph_node_inside_block():
    templates = {
        "block.j2": "{{ content }}",
        "text.j2": "{{ value }}",
        "paragraph.j2": (
            "{{ content }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = BlockNode(
        subtype="section",
        children=[
            ParagraphNode(
                children=[TextNode("Just some text")],
                args=["arg1", "arg2"],
                kwargs={"key1": "value1"},
                tags=["tag1", "tag2"],
            ),
        ],
    )

    result = visitor.visit(node)

    assert result == "Just some text - arg1,arg2 - key1:value1 - tag1,tag2"



from mau.lexers.text_lexer import TextLexer
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_collect_macro_arguments_single_argument():
    source = "(value1)"

    parser = init_parser(source)

    assert parser._collect_macro_args() == "value1"


def test_collect_macro_arguments_multiple_arguments():
    source = "(value1,value2)"

    parser = init_parser(source)

    assert parser._collect_macro_args() == "value1,value2"


def test_collect_macro_arguments_single_argument_with_quotes():
    source = '("value1")'

    parser = init_parser(source)

    assert parser._collect_macro_args() == '"value1"'


def test_collect_macro_arguments_single_argument_with_quotes_and_parenthesis():
    source = '("value1()")'

    parser = init_parser(source)

    assert parser._collect_macro_args() == '"value1()"'


def test_collect_macro_arguments_single_argument_with_parenthesis():
    source = "(value1())"

    parser = init_parser(source)

    assert parser._collect_macro_args() == "value1("


def test_collect_macro_arguments_multiple_argument_with_quotes_and_parenthesis():
    source = '("value1()",value2,value3)'

    parser = init_parser(source)

    assert parser._collect_macro_args() == '"value1()",value2,value3'


def test_collect_macro_arguments_multiple_argument_with_escaped_quotes():
    source = r"(\"value2,value3)"

    parser = init_parser(source)

    assert parser._collect_macro_args() == r"\"value2,value3"



from mau.lexers.text_lexer import TextLexer
from mau.nodes.inline import StyleNode, TextNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_underscore():
    source = "_Some text_"

    expected = [
        StyleNode(
            value="underscore",
            children=[
                TextNode("Some text"),
            ],
        ),
    ]

    assert runner(source).nodes == expected


def test_star():
    source = "*Some text*"

    expected = [
        StyleNode(
            value="star",
            children=[
                TextNode("Some text"),
            ],
        ),
    ]

    assert runner(source).nodes == expected


def test_caret():
    source = "^Some text^"

    expected = [
        StyleNode(
            value="caret",
            children=[
                TextNode("Some text"),
            ],
        ),
    ]

    assert runner(source).nodes == expected


def test_tilde():
    source = "~Some text~"

    expected = [
        StyleNode(
            value="tilde",
            children=[
                TextNode("Some text"),
            ],
        ),
    ]

    assert runner(source).nodes == expected


def test_style_within_style():
    source = "_*Words with two styles*_"

    expected = [
        StyleNode(
            value="underscore",
            children=[
                StyleNode(
                    value="star",
                    children=[TextNode("Words with two styles")],
                )
            ],
        )
    ]

    assert runner(source).nodes == expected


def test_double_style_cancels_itself():
    source = "__Text__"

    expected = [
        StyleNode(value="underscore"),
        TextNode("Text"),
        StyleNode(value="underscore"),
    ]

    assert runner(source).nodes == expected


def test_mix_text_and_styles():
    source = "Some text _and style_ and *more style* here"

    expected = [
        TextNode("Some text "),
        StyleNode(
            value="underscore",
            children=[
                TextNode("and style"),
            ],
        ),
        TextNode(" and "),
        StyleNode(
            value="star",
            children=[
                TextNode("more style"),
            ],
        ),
        TextNode(" here"),
    ]

    assert runner(source).nodes == expected


def test_unclosed_style():
    source = "_Text"

    expected = [
        TextNode("_Text"),
    ]

    assert runner(source).nodes == expected



import hashlib
import re

from mau.nodes.toc import TocEntryNode, TocNode


class TocManager:
    def __init__(self, parser):
        # This list contains the headers
        # found parsing a document
        self.headers = []

        # This list contains all the ToC entries
        # that will be shown by a toc command.
        self.tocnodes = []

        # This is the list of ::toc commands
        # that need to be updated once the ToC
        # has been processed
        self.command_nodes = []

        # This is the parser that contains the manager
        self.parser = parser

    def add_header_node(self, node):
        self.headers.append(node)

    def create_toc_node(self, subtype, args, kwargs, tags):
        # This creates an empty TocNode
        # that will be stored in the parser
        # and in the command_nodes list.
        # The method process_footnotes will
        # eventually update all these nodes
        # with the right entries.
        node = TocNode(
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )

        self.command_nodes.append(node)
        self.parser.save(node)

    def process_toc(self):
        toc = TocNode(children=create_toc(self.headers))

        for node in self.command_nodes:
            node.children = create_toc(
                self.headers, exclude_tag=node.kwargs.get("exclude_tag")
            )

        return toc

    def update(self, other):
        self.headers.extend(other.headers)
        self.command_nodes.extend(other.command_nodes)


def create_toc(headers, exclude_tag=None):
    # Create the TOC from the list of headers.

    nodes = []
    latest_by_level = {}

    for header_node in headers:
        # This is the current node
        node = TocEntryNode(
            value=header_node.value,
            anchor=header_node.anchor,
            tags=header_node.tags,
        )
        level = int(header_node.level)

        # This collects the latest node added with a given level
        latest_by_level[level] = node

        try:
            # Simplest case, add it to the latest one
            # with a level just 1 step lower
            latest_by_level[level - 1].children.append(node)
        except KeyError:
            # Find all the latest ones added with a level lower than this
            latest = [latest_by_level.get(i, None) for i in range(1, level)]

            # Get the children list of each one, plus nodes for the root
            children = [nodes] + [i.children for i in latest if i is not None]

            # Get the nearest one and append to that
            children[-1].append(node)

    if exclude_tag:
        nodes = exclude_children(nodes, exclude_tag)

    return nodes


def header_anchor(text, level):
    """
    Return a sanitised anchor for a header.
    """

    # Everything lowercase
    sanitised_text = text.lower()

    # Get only letters, numbers, dashes, spaces, and dots
    sanitised_text = "".join(re.findall("[a-z0-9-\\. ]+", sanitised_text))

    # Remove multiple spaces
    sanitised_text = "-".join(sanitised_text.split())

    hashed_value = hashlib.md5(f"{level} {text}".encode("utf-8")).hexdigest()[:4]

    return f"{sanitised_text}-{hashed_value}"


def exclude_children(children, exclude_tag):
    valid_children = [i for i in children if exclude_tag not in i.tags]

    for entry in valid_children:
        entry.children = exclude_children(entry.children, exclude_tag)

    return valid_children



from mau.nodes.header import HeaderNode


def test_header_node():
    node = HeaderNode(
        "somevalue",
        "somelevel",
        "someanchor",
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )

    assert node.value == "somevalue"
    assert node.level == "somelevel"
    assert node.anchor == "someanchor"
    assert node.args == ["value1", "value2"]
    assert node.tags == ["tag1", "tag2"]
    assert node.kwargs == {"key1": "text1", "key2": "text2"}
    assert node.node_type == "header"
    assert node == HeaderNode(
        "somevalue",
        "somelevel",
        "someanchor",
        args=["value1", "value2"],
        tags=["tag1", "tag2"],
        kwargs={"key1": "text1", "key2": "text2"},
    )



from mau.nodes.nodes import Node


class ParagraphNode(Node):
    """A paragraph."""

    node_type = "paragraph"

    def __init__(
        self,
        title=None,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.title = title

    def _custom_dict(self):
        return {
            "title": self.title,
        }



from mau.nodes.nodes import Node


class ListItemNode(Node):
    """An entry in a list."""

    node_type = "list_item"

    def __init__(
        self,
        level,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.level = level

    def _custom_dict(self):
        return {
            "level": self.level,
        }


class ListNode(Node):
    """A list."""

    node_type = "list"

    def __init__(
        self,
        ordered,
        main_node=False,
        start=1,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.ordered = ordered
        self.main_node = main_node
        self.start = start

    def _custom_dict(self):
        return {
            "ordered": self.ordered,
            "main_node": self.main_node,
            "start": self.start,
        }



import pytest
from mau.environment.environment import Environment
from mau.errors import MauErrorException
from mau.lexers.text_lexer import TextLexer
from mau.nodes.inline import SentenceNode, StyleNode, TextNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_control_if_true():
    environment = Environment({"flag": True})

    source = '[@if:flag:&true]("TRUE", "FALSE")'

    expected = [
        SentenceNode(
            children=[TextNode("TRUE")],
        ),
    ]

    assert runner(source, environment=environment).nodes == expected


def test_control_if_false():
    environment = Environment({"flag": True})

    source = '[@if:flag:&false]("TRUE", "FALSE")'

    expected = [
        SentenceNode(
            children=[TextNode("FALSE")],
        ),
    ]

    assert runner(source, environment=environment).nodes == expected


def test_control_if_equal():
    environment = Environment({"flag": "42"})

    source = '[@if:flag:=42]("TRUE", "FALSE")'

    expected = [
        SentenceNode(
            children=[TextNode("TRUE")],
        ),
    ]

    assert runner(source, environment=environment).nodes == expected


def test_control_if_not_equal():
    environment = Environment({"flag": "42"})

    source = '[@if:flag:!=42]("TRUE", "FALSE")'

    expected = [
        SentenceNode(
            children=[TextNode("FALSE")],
        ),
    ]

    assert runner(source, environment=environment).nodes == expected


def test_control_wrong_name_format():
    source = '[@if:flag]("TRUE", "FALSE")'

    with pytest.raises(MauErrorException):
        runner(source)


def test_control_unsupported_operator():
    source = '[@something:flag:&true]("TRUE", "FALSE")'

    with pytest.raises(MauErrorException):
        runner(source)


def test_control_undefined_variable():
    source = '[@if:flag:&true]("TRUE", "FALSE")'

    with pytest.raises(MauErrorException):
        runner(source)


def test_control_invalid_boolean():
    environment = Environment({"flag": "42"})

    source = '[@if:flag:&something]("TRUE", "FALSE")'

    with pytest.raises(MauErrorException):
        runner(source, environment)


def test_control_invalid_test():
    environment = Environment({"flag": "42"})

    source = '[@if:flag:-something]("TRUE", "FALSE")'

    with pytest.raises(MauErrorException):
        runner(source, environment)


def test_macro_ifeval_true():
    environment = Environment(
        {
            "flag": True,
            "underscore": "_sometext_",
            "star": "*othertext*",
        }
    )

    source = "[@ifeval:flag:&true](underscore, star)"

    expected = [
        SentenceNode(
            children=[
                StyleNode(
                    value="underscore",
                    children=[
                        TextNode("sometext"),
                    ],
                )
            ],
        ),
    ]

    assert runner(source, environment=environment).nodes == expected


def test_macro_ifeval_false():
    environment = Environment(
        {
            "flag": True,
            "underscore": "_sometext_",
            "star": "*othertext*",
        }
    )

    source = "[@ifeval:flag:&false](underscore, star)"

    expected = [
        SentenceNode(
            children=[
                StyleNode(
                    value="star",
                    children=[
                        TextNode("othertext"),
                    ],
                )
            ],
        ),
    ]

    assert runner(source, environment=environment).nodes == expected


def test_macro_ifeval_false_is_not_evaluated():
    environment = Environment(
        {
            "flag": True,
            "style": "_sometext_",
            "header": "[header](notexists)",
        }
    )

    source = "[@ifeval:flag:&true](style, header)"

    expected = [
        SentenceNode(
            children=[
                StyleNode(
                    value="underscore",
                    children=[
                        TextNode("sometext"),
                    ],
                )
            ],
        ),
    ]

    assert runner(source, environment=environment).nodes == expected


def test_macro_ifeval_true_is_not_evaluated():
    environment = Environment(
        {
            "flag": True,
            "style": "_sometext_",
            "header": "[header](notexists)",
        }
    )

    source = "[@ifeval:flag:&false](header, style)"

    expected = [
        SentenceNode(
            children=[
                StyleNode(
                    value="underscore",
                    children=[
                        TextNode("sometext"),
                    ],
                )
            ],
        ),
    ]

    assert runner(source, environment=environment).nodes == expected


def test_macro_ifeval_true_not_defined():
    environment = Environment(
        {
            "flag": True,
            "star": "*othertext*",
        }
    )

    source = "[@ifeval:flag:&true](underscore, star)"

    with pytest.raises(MauErrorException):
        runner(source, environment)


def test_macro_ifeval_false_not_defined():
    environment = Environment(
        {
            "flag": True,
            "underscore": "_sometext_",
        }
    )

    source = "[@ifeval:flag:&false](underscore, star)"

    with pytest.raises(MauErrorException):
        runner(source, environment)



from mau.environment.environment import Environment
from mau.nodes.inline import TextNode
from mau.nodes.source import CalloutNode, CalloutsEntryNode, SourceNode
from mau.visitors.jinja_visitor import JinjaVisitor


def test_source_node():
    templates = {
        "text.j2": "{{ value }}",
        "callout.j2": "",
        "callouts_entry.j2": "{{ marker }} - {{ value }}",
        "source.custom.j2": "The custom template",
        "source.j2": "The default template",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = SourceNode(
        language="somelang",
        code=[
            TextNode("import sys"),
            TextNode("import: os"),
            TextNode(""),
            TextNode('print(os.environ["HOME"])'),
        ],
        markers=[CalloutNode(1, "imp"), CalloutNode(3, "env")],
        callouts=[
            CalloutsEntryNode("imp", "This is an import"),
            CalloutsEntryNode("env", "Environment variables are paramount"),
        ],
        subtype="custom",
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == "The custom template"



from mau.environment.environment import Environment
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.toc import TocEntryNode, TocNode
from mau.visitors.base_visitor import BaseVisitor


def test_toc_node():
    visitor = BaseVisitor(Environment())

    node = TocNode(
        children=[
            TocEntryNode(
                value=SentenceNode(children=[TextNode("Header 1")]),
                anchor="header-1",
                children=[
                    TocEntryNode(
                        value=SentenceNode(children=[TextNode("Header 1.1")]),
                        anchor="header-1-1",
                    ),
                ],
            ),
            TocEntryNode(
                value=SentenceNode(children=[TextNode("Header 2")]),
                anchor="header-2",
            ),
        ],
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "toc",
            "entries": [
                {
                    "data": {
                        "anchor": "header-1",
                        "children": [
                            {
                                "data": {
                                    "anchor": "header-1-1",
                                    "children": [],
                                    "tags": [],
                                    "type": "toc_entry",
                                    "value": {
                                        "data": {
                                            "type": "sentence",
                                            "subtype": None,
                                            "args": [],
                                            "kwargs": {},
                                            "tags": [],
                                            "content": [
                                                {
                                                    "data": {
                                                        "type": "text",
                                                        "value": "Header 1.1",
                                                        "subtype": None,
                                                        "args": [],
                                                        "kwargs": {},
                                                        "tags": [],
                                                    }
                                                }
                                            ],
                                        }
                                    },
                                }
                            }
                        ],
                        "tags": [],
                        "type": "toc_entry",
                        "value": {
                            "data": {
                                "type": "sentence",
                                "subtype": None,
                                "args": [],
                                "kwargs": {},
                                "tags": [],
                                "content": [
                                    {
                                        "data": {
                                            "type": "text",
                                            "value": "Header 1",
                                            "subtype": None,
                                            "args": [],
                                            "kwargs": {},
                                            "tags": [],
                                        }
                                    }
                                ],
                            }
                        },
                    }
                },
                {
                    "data": {
                        "type": "toc_entry",
                        "anchor": "header-2",
                        "children": [],
                        "tags": [],
                        "value": {
                            "data": {
                                "type": "sentence",
                                "subtype": None,
                                "args": [],
                                "kwargs": {},
                                "tags": [],
                                "content": [
                                    {
                                        "data": {
                                            "type": "text",
                                            "value": "Header 2",
                                            "subtype": None,
                                            "args": [],
                                            "kwargs": {},
                                            "tags": [],
                                        }
                                    }
                                ],
                            }
                        },
                    }
                },
            ],
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
        }
    }



from mau.lexers.main_lexer import MainLexer
from mau.nodes.block import BlockNode
from mau.nodes.content import ContentNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.nodes import Node
from mau.parsers.main_parser import MainParser
from mau.text_buffer.context import Context

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_initial_title():
    source = """
    """

    parser = runner(source)

    assert parser.title == (None, None)


def test_push_title():
    source = """
    """

    parser = runner(source)

    parser._push_title("Just a title", Context(42, 128, "main", ".Just a title"))

    assert parser.title == (
        "Just a title",
        Context(42, 128, "main", ".Just a title"),
    )


def test_pop_title():
    source = """
    """

    node = Node()

    parser = runner(source)

    parser._push_title("Just a title", Context(42, 128, "main", ".Just a title"))

    title_node = parser._pop_title(node)

    assert title_node == SentenceNode(
        children=[
            TextNode("Just a title"),
        ]
    )
    assert title_node.parent == node
    assert title_node.parent_position == "title"


def test_parse_title():
    source = """
    .Just a title
    """

    parser = runner(source)

    assert parser.title == (
        "Just a title",
        Context(1, 0, "main", ".Just a title"),
    )


def test_block_uses_title():
    source = """
    .Just a title
    ----
    ----
    """

    parser = runner(source)

    assert parser.nodes == [
        BlockNode(
            subtype=None,
            classes=[],
            title=SentenceNode(
                children=[
                    TextNode("Just a title"),
                ]
            ),
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]

    assert parser.title == (None, None)


def test_include_content_uses_title():
    source = """
    .Just a title
    << type:uri1
    """

    parser = runner(source)

    assert parser.nodes == [
        ContentNode(
            content_type="type",
            uris=["uri1"],
            title=SentenceNode(
                children=[
                    TextNode("Just a title"),
                ]
            ),
        )
    ]

    assert parser.title == (None, None)



import pytest
from mau.environment.environment import Environment
from mau.errors import MauErrorException
from mau.nodes.footnotes import FootnoteNode
from mau.nodes.header import HeaderNode
from mau.nodes.inline import SentenceNode, StyleNode, TextNode, VerbatimNode
from mau.nodes.macros import (
    MacroClassNode,
    MacroHeaderNode,
    MacroImageNode,
    MacroLinkNode,
    MacroNode,
)
from mau.visitors.jinja_visitor import JinjaVisitor


def test_default_values():
    visitor = JinjaVisitor(Environment())

    assert visitor.default_templates == Environment()
    assert visitor.environment_options == {}
    assert visitor.extension == "j2"
    assert visitor.toc is None
    assert visitor.footnotes is None

    assert visitor.templates == Environment()


def test_custom_templates():
    templates = {"key": "value"}

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")

    visitor = JinjaVisitor(environment)

    assert visitor.templates.asdict() == templates


def test_custom_templates_with_default_templates():
    class MyJinjaVisitor(JinjaVisitor):
        default_templates = {"defaultkey": "defaultvalue"}

    templates = {"key": "value"}
    merged_templates = Environment({"defaultkey": "defaultvalue", "key": "value"})

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")

    visitor = MyJinjaVisitor(environment)

    assert visitor.templates == merged_templates


def test_no_templates():
    templates = {}

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = TextNode("Just some text.")

    with pytest.raises(MauErrorException):
        visitor.visit(node)


def test_inline_text_node():
    templates = {
        "text.j2": "{{ value }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = TextNode("Just some text.")

    result = visitor.visit(node)

    assert result == "Just some text."


def test_inline_verbatim_node():
    templates = {
        "verbatim.j2": "{{ value }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = VerbatimNode("Just some text.")

    result = visitor.visit(node)

    assert result == "Just some text."


def test_inline_style_node_star():
    templates = {
        "text.j2": "{{ value }}",
        "style.star.j2": "*{{ content }}*",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = StyleNode("star", children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == "*Just some text.*"


def test_inline_style_node_underscore():
    templates = {
        "text.j2": "{{ value }}",
        "style.underscore.j2": "_{{ content }}_",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = StyleNode("underscore", children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == "_Just some text._"


def test_inline_style_node_tilde():
    templates = {
        "text.j2": "{{ value }}",
        "style.tilde.j2": "~{{ content }}~",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = StyleNode("tilde", children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == "~Just some text.~"


def test_inline_style_node_caret():
    templates = {
        "text.j2": "{{ value }}",
        "style.caret.j2": "^{{ content }}^",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = StyleNode("caret", children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == "^Just some text.^"


def test_inline_macro_node():
    templates = {
        "text.j2": "{{ value }}",
        "macro.j2": (
            "{{ name }} - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = MacroNode("somename", args=["arg1", "arg2"], kwargs={"key1": "value1"})

    result = visitor.visit(node)

    assert result == "somename - arg1,arg2 - key1:value1"


def test_inline_footnote_node():
    templates = {
        "text.j2": "{{ value }}",
        "footnote.j2": (
            "{{ content }} - {{ number }} - "
            "{{ reference_anchor }} - {{ content_anchor }}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = FootnoteNode(
        children=[TextNode("Just some text.")],
        number="5",
        reference_anchor="someanchor",
        content_anchor="someanchor-def",
    )

    result = visitor.visit(node)

    assert result == "Just some text. - 5 - someanchor - someanchor-def"


def test_inline_class_node():
    templates = {
        "text.j2": "{{ value }}",
        "macro.class.j2": "{{ classes | join(',') }} - {{ content }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = MacroClassNode(["class1", "class2"], children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == "class1,class2 - Just some text."


def test_inline_link_node():
    templates = {
        "text.j2": "{{ value }}",
        "macro.link.j2": "{{ target }} - {{ content }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = MacroLinkNode(target="sometarget", children=[TextNode("sometext")])

    result = visitor.visit(node)

    assert result == "sometarget - sometext"


def test_inline_header_node():
    templates = {
        "text.j2": "{{ value }}",
        "sentence.j2": "{{ content }}",
        "macro.header.j2": "#{{ header.anchor }} - {{ content }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    header_node = HeaderNode(
        value=SentenceNode(children=[TextNode("Header")]),
        level="2",
        anchor="XXXXXX",
        kwargs={"id": "someid"},
    )

    node = MacroHeaderNode(
        header_id="someid", header=header_node, children=[TextNode("sometext")]
    )

    result = visitor.visit(node)

    assert result == "#XXXXXX - sometext"


def test_inline_image_node():
    templates = {
        "text.j2": "{{ value }}",
        "macro.image.j2": "{{ uri }} - {{ alt_text }} - {{ width }}x{{ height }}",
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")
    visitor = JinjaVisitor(environment)

    node = MacroImageNode(uri="someuri", alt_text="sometext", width="100", height="400")

    result = visitor.visit(node)

    assert result == "someuri - sometext - 100x400"



import pytest
from mau.errors import MauErrorException
from mau.lexers.main_lexer import MainLexer
from mau.nodes.block import BlockGroupNode, BlockNode
from mau.nodes.inline import TextNode
from mau.nodes.paragraph import ParagraphNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_engine_group():
    source = """
    [*sometype1, engine=group, group=somegroup, position=left]
    ----
    Block 1
    ----
    
    [*sometype2, engine=group, group=somegroup, position=right]
    ----
    Block 2
    ----
    """

    par = runner(source)

    assert par.nodes == []

    assert par.grouped_blocks == {
        "somegroup": {
            "left": BlockNode(
                subtype="sometype1",
                children=[
                    ParagraphNode(
                        children=[
                            TextNode("Block 1"),
                        ]
                    ),
                ],
                secondary_children=[],
                classes=[],
                title=None,
                engine="group",
                preprocessor="none",
                args=[],
                kwargs={},
            ),
            "right": BlockNode(
                subtype="sometype2",
                children=[
                    ParagraphNode(
                        children=[
                            TextNode("Block 2"),
                        ]
                    ),
                ],
                secondary_children=[],
                classes=[],
                title=None,
                engine="group",
                preprocessor="none",
                args=[],
                kwargs={},
            ),
        }
    }


def test_engine_group_duplicate_position():
    source = """
    [*sometype1, engine=group, group=somegroup, position=left]
    ----
    Block 1
    ----
    
    [*sometype2, engine=group, group=somegroup, position=left]
    ----
    Block 2
    ----
    """

    with pytest.raises(MauErrorException):
        runner(source)


def test_engine_group_command_group():
    source = """
    [*sometype1, engine=group, group=somegroup, position=left]
    ----
    Block 1
    ----
    
    [*sometype2, engine=group, group=somegroup, position=right]
    ----
    Block 2
    ----

    ::blockgroup:somegroup
    """

    par = runner(source)

    assert par.nodes == [
        BlockGroupNode(
            group_name="somegroup",
            subtype=None,
            children=[],
            group={
                "left": BlockNode(
                    subtype="sometype1",
                    children=[
                        ParagraphNode(
                            children=[
                                TextNode("Block 1"),
                            ]
                        ),
                    ],
                    secondary_children=[],
                    classes=[],
                    title=None,
                    engine="group",
                    preprocessor="none",
                    args=[],
                    kwargs={},
                ),
                "right": BlockNode(
                    subtype="sometype2",
                    children=[
                        ParagraphNode(
                            children=[
                                TextNode("Block 2"),
                            ]
                        ),
                    ],
                    secondary_children=[],
                    classes=[],
                    title=None,
                    engine="group",
                    preprocessor="none",
                    args=[],
                    kwargs={},
                ),
            },
            title=None,
            args=[],
            kwargs={},
        )
    ]

    assert par.grouped_blocks == {}

    block_group = par.nodes[0]
    block_left = block_group.group["left"]
    block_right = block_group.group["right"]

    assert block_left.parent == block_group
    assert block_left.parent_position == "left"

    assert block_right.parent == block_group
    assert block_right.parent_position == "right"


def test_engine_group_command_group_invalid_group():
    source = """
    ::blockgroup:somegroup
    """

    with pytest.raises(MauErrorException):
        runner(source)



from mau.lexers.main_lexer import MainLexer
from mau.parsers.main_parser import MainParser

from tests.helpers import parser_runner_factory

mainrunner = parser_runner_factory(MainLexer, MainParser)


def test_parent_block():
    source = """
    ----
    Content
    ----
    """

    nodes = mainrunner(source).nodes

    block_node = nodes[0]
    paragraph_node = block_node.children[0]

    assert paragraph_node.parent is block_node


def test_parent_paragraph():
    source = "Some text _and style_ and *more style* here"

    nodes = mainrunner(source).nodes

    paragraph_node = nodes[0]
    text_node = paragraph_node.children[0]
    style_node = paragraph_node.children[1]

    assert text_node.parent is paragraph_node
    assert style_node.parent is paragraph_node



from mau.environment.environment import Environment
from mau.nodes.inline import TextNode
from mau.nodes.page import DocumentNode, HorizontalRuleNode
from mau.nodes.paragraph import ParagraphNode
from mau.visitors.base_visitor import BaseVisitor


def test_horizontal_rule_node():
    visitor = BaseVisitor(Environment())

    node = HorizontalRuleNode(
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "horizontal_rule",
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
            "subtype": None,
        }
    }


def test_document_node():
    visitor = BaseVisitor(Environment())

    node = DocumentNode(
        children=[
            ParagraphNode(
                children=[TextNode("Just some text")],
                args=["arg1", "arg2"],
                kwargs={"key1": "value1"},
                tags=["tag1", "tag2"],
            )
        ],
        args=["arg3", "arg4"],
        kwargs={"key2": "value2"},
        tags=["tag3", "tag4"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "document",
            "content": [
                {
                    "data": {
                        "type": "paragraph",
                        "content": [
                            {
                                "data": {
                                    "type": "text",
                                    "value": "Just some text",
                                    "args": [],
                                    "kwargs": {},
                                    "subtype": None,
                                    "tags": [],
                                }
                            }
                        ],
                        "title": {},
                        "subtype": None,
                        "args": ["arg1", "arg2"],
                        "kwargs": {"key1": "value1"},
                        "tags": ["tag1", "tag2"],
                    }
                }
            ],
            "args": ["arg3", "arg4"],
            "kwargs": {"key2": "value2"},
            "tags": ["tag3", "tag4"],
            "subtype": None,
        }
    }



from mau.lexers.text_lexer import TextLexer
from mau.nodes.inline import TextNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_dollar():
    source = "$Many different words$"

    expected = [
        TextNode("Many different words"),
    ]

    assert runner(source).nodes == expected


def test_percent():
    source = "%Many different words%"

    expected = [
        TextNode("Many different words"),
    ]

    assert runner(source).nodes == expected


def test_dollar_can_escape_percent():
    source = "$Many %different% words$"

    expected = [
        TextNode("Many %different% words"),
    ]

    assert runner(source).nodes == expected


def test_dollar_can_escape_backtick():
    source = "$Many `different` words$"

    expected = [
        TextNode("Many `different` words"),
    ]

    assert runner(source).nodes == expected


def test_percent_can_escape_dollar():
    source = "%Many $different$ words%"

    expected = [
        TextNode("Many $different$ words"),
    ]

    assert runner(source).nodes == expected


def test_percent_can_escape_backtick():
    source = "%Many `different` words%"

    expected = [
        TextNode("Many `different` words"),
    ]

    assert runner(source).nodes == expected


def test_dollar_escapes_style():
    source = "$Many _different_ words$"

    expected = [
        TextNode("Many _different_ words"),
    ]

    assert runner(source).nodes == expected


def test_percent_escapes_style():
    source = "%Many _different_ words%"

    expected = [
        TextNode("Many _different_ words"),
    ]

    assert runner(source).nodes == expected


def test_escaped_dollar():
    source = r"$\$$"

    expected = [
        TextNode("$"),
    ]

    assert runner(source).nodes == expected


def test_escaped_percent():
    source = r"%\%%"

    expected = [
        TextNode("%"),
    ]

    assert runner(source).nodes == expected


def test_open_dollar():
    source = r"$Many different words"

    expected = [
        TextNode("$Many different words"),
    ]

    assert runner(source).nodes == expected


def test_open_percent():
    source = r"%Many different words"

    expected = [
        TextNode("%Many different words"),
    ]

    assert runner(source).nodes == expected



from mau.nodes.page import Node


class ContentNode(Node):
    """Content included in the page.

    This represents generic content included in the page.

    Arguments:
        content_type: the type of content
        title: caption of the image
        args: unnamed arguments
        kwargs: named arguments
        tags: tags
    """

    node_type = "content"

    def __init__(
        self,
        content_type,
        uris,
        title=None,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.content_type = content_type
        self.title = title
        self.uris = uris

    def _custom_dict(self):
        return {
            "content_type": self.content_type,
            "title": self.title,
            "uris": self.uris,
        }


class ContentImageNode(Node):
    """An image included in the page."""

    node_type = "content_image"

    def __init__(
        self,
        uri,
        alt_text=None,
        classes=None,
        title=None,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.uri = uri
        self.title = title
        self.alt_text = alt_text
        self.classes = classes

    def _custom_dict(self):
        return {
            "uri": self.uri,
            "title": self.title,
            "alt_text": self.alt_text,
            "classes": self.classes,
        }



import hashlib

from mau.nodes.footnotes import FootnotesNode


class FootnotesManager:
    def __init__(self, parser):
        # This dictionary containes the footnotes created
        # in the text through a macro.
        self.mentions = {}

        # This dictionary contains the content of each
        # footnote created through blocks.
        # This is a helper dictionary that will be merged
        # with self.footnotes once the parsing is completed.
        self.data = {}

        # This list contains all the footnote entries
        # that will be shown by a footnotes command.
        self.footnotes = []

        # This is the list of ::footnotes commands
        # that need to be updated once footnotes
        # have been processed
        self.command_nodes = []

        # This is the parser that contains the manager
        self.parser = parser

    def create_node(self, subtype, args, kwargs, tags):
        # This creates an empty FootnotesNode
        # that will be stored in the parser
        # and in the command_nodes list.
        # The method process_footnotes will
        # eventually update all these nodes
        # with the right entries.
        node = FootnotesNode(
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )

        self.command_nodes.append(node)
        self.parser.save(node)

    def add_data(self, name, content):
        self.data[name] = content

    def process_footnotes(self):
        self.footnotes = create_footnotes(
            self.mentions,
            self.data,
        )

        for node in self.command_nodes:
            # create_footnotes is called
            # multiple times to create new
            # entries for each command.
            # If we don't do it, all commands
            # will share the same entries and in case
            # of multiple ::footnotes: the parent
            # of all entries will be the last
            # command footnotes.
            node.add_children(
                create_footnotes(
                    self.mentions,
                    self.data,
                )
            )

            # node.children = create_footnotes(
            #     self.mentions,
            #     self.data,
            # )

            # for child in node.children:
            #     child.parent = node

    def update(self, other):
        self.update_mentions(other.mentions)
        self.data.update(other.data)
        self.command_nodes.extend(other.command_nodes)

    def update_mentions(self, mentions):
        # Retrieve the footnotes
        # The format of this dictionary is
        # {name: node}
        text_footnotes = set(mentions.keys())
        existing_footnotes = set(self.mentions.keys())
        duplicate_footnotes = set.intersection(existing_footnotes, text_footnotes)

        if duplicate_footnotes:
            duplicates_list = ", ".join(duplicate_footnotes)
            self.parser._error(f"Duplicate footnotes detected: {duplicates_list}")

        self.mentions.update(mentions)


def footnote_anchor(content):
    return hashlib.md5(str(content).encode("utf-8")).hexdigest()[:8]


def create_footnotes(footnote_mentions, footnote_data):
    footnotes = []

    for num, footnote in enumerate(footnote_mentions.values(), start=1):
        footnote.number = num

    for key, footnote in footnote_mentions.items():
        data = footnote_data[key]
        footnote.children = data
        anchor = footnote_anchor(footnote.children)

        footnote.reference_anchor = f"ref-footnote-{footnote.number}-{anchor}"
        footnote.content_anchor = f"cnt-footnote-{footnote.number}-{anchor}"

        footnotes.append(footnote.to_entry())

    return footnotes



import pytest
from mau.environment.environment import Environment
from mau.errors import MauErrorException
from mau.lexers.main_lexer import MainLexer
from mau.nodes.block import BlockNode
from mau.nodes.content import ContentNode
from mau.nodes.header import HeaderNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.nodes.paragraph import ParagraphNode
from mau.parsers.main_parser import MainParser
from mau.text_buffer.context import Context

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_initial_control():
    source = """
    """

    parser = runner(source)

    assert parser.control == (None, None, None)


def test_push_control():
    source = """
    """

    parser = runner(source)

    parser._push_control(
        "operator",
        "statement",
        Context(42, 128, "main", "@operator:statement"),
    )

    assert parser.control == (
        "operator",
        "statement",
        Context(42, 128, "main", "@operator:statement"),
    )


def test_pop_control_default():
    source = """
    """

    parser = runner(source)

    assert parser._pop_control() is True


def test_pop_control_equal():
    source = """
    :flag:42
    """

    parser = runner(source)

    parser._push_control("if", "flag:=42", Context(42, 128, "main", "@if:flag"))

    assert parser._pop_control() is True


def test_pop_control_not_equal():
    source = """
    :flag:42
    """

    parser = runner(source)

    parser._push_control("if", "flag:!=42", Context(42, 128, "main", "@if:flag"))

    assert parser._pop_control() is False


def test_pop_control_boolean_true():
    source = """
    :+flag:
    """

    parser = runner(source)

    parser._push_control("if", "flag:&true", Context(42, 128, "main", "@if:flag"))

    assert parser._pop_control() is True


def test_pop_control_boolean_false():
    source = """
    :+flag:
    """

    parser = runner(source)

    parser._push_control("if", "flag:&false", Context(42, 128, "main", "@if:flag"))

    assert parser._pop_control() is False


def test_pop_control_boolean_invalid():
    source = """
    :+flag:
    """

    parser = runner(source)

    parser._push_control("if", "flag:&something", Context(42, 128, "main", "@if:flag"))

    with pytest.raises(MauErrorException):
        parser._pop_control()


def test_pop_control_not_if():
    source = """
    """

    parser = runner(source)

    parser._push_control(
        "something", "flag:&something", Context(42, 128, "main", "@if:flag")
    )

    with pytest.raises(MauErrorException):
        parser._pop_control()


def test_pop_control_invalid_statement_format():
    source = """
    """

    parser = runner(source)

    parser._push_control("if", "flag=42", Context(42, 128, "main", "@if:flag"))

    with pytest.raises(MauErrorException):
        parser._pop_control()


def test_pop_control_variable_not_defined():
    source = """
    """

    parser = runner(source)

    parser._push_control("if", "flag:=42", Context(42, 128, "main", "@if:flag"))

    with pytest.raises(MauErrorException):
        parser._pop_control()


def test_pop_control_test_not_supported():
    source = """
    :+flag:
    """

    parser = runner(source)

    parser._push_control("if", "flag:*42", Context(42, 128, "main", "@if:flag"))

    with pytest.raises(MauErrorException):
        parser._pop_control()


def test_parse_control():
    source = """
    @operator:variable:test
    """

    parser = runner(source)

    assert parser.control == (
        "operator",
        "variable:test",
        Context(1, 0, "main", "@operator:variable:test"),
    )


def test_block_uses_control_true():
    source = """
    :+flag:
    
    @if:flag:&true
    ----
    ----
    """

    parser = runner(source)

    assert parser.nodes == [
        BlockNode(
            subtype=None,
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]

    assert parser.control == (None, None, None)


def test_block_uses_control_false():
    source = """
    :-flag:
    
    @if:flag:&true
    ----
    ----
    """

    parser = runner(source)

    assert parser.nodes == []

    assert parser.control == (None, None, None)


def test_include_content_uses_control_true():
    source = """
    :+flag:
    
    @if:flag:&true
    << ctype:/path/to/it
    """

    parser = runner(source)

    assert parser.nodes == [
        ContentNode(
            content_type="ctype",
            uris=["/path/to/it"],
            title=None,
        )
    ]

    assert parser.control == (None, None, None)


def test_include_content_uses_control_false():
    source = """
    :-flag:
    
    @if:flag:&true
    << ctype:/path/to/it
    """

    parser = runner(source)

    assert parser.nodes == []

    assert parser.control == (None, None, None)


def test_header_uses_control_true():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: "XXXXXX"
    )

    source = """
    :+flag:
    
    @if:flag:&true
    = Header
    """

    parser = runner(source, environment)

    assert parser.nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header")]),
            level="1",
            anchor="XXXXXX",
        )
    ]

    assert parser.control == (None, None, None)


def test_header_uses_control_false():
    source = """
    :-flag:
    
    @if:flag:&true
    = Header
    """

    parser = runner(source)

    assert parser.nodes == []

    assert parser.control == (None, None, None)


def test_paragraph_uses_control_true():
    source = """
    :+flag:
    
    @if:flag:&true
    This is a paragraph.
    """

    parser = runner(source)

    assert parser.nodes == [ParagraphNode(children=[TextNode("This is a paragraph.")])]

    assert parser.control == (None, None, None)


def test_paragraph_uses_control_false():
    source = """
    :-flag:
    
    @if:flag:&true
    This is a paragraph.
    """

    parser = runner(source)

    assert parser.nodes == []

    assert parser.control == (None, None, None)



import hashlib


class Node:
    node_type = "node"

    def __init__(
        self,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        self.parent = parent
        self.parent_position = parent_position
        self.subtype = subtype
        self.children = children or []
        self.args = args or []
        self.kwargs = kwargs or {}
        self.tags = tags or []

    def _custom_dict(self):
        return {}

    def add_children(self, children):
        self.children.extend(children)
        for child in children:
            child.parent = self

    def asdict(self):
        base = {
            # Parent is excluded to avoid
            # having to deal with recursion
            "type": self.node_type,
            "subtype": self.subtype,
            "children": [i.asdict() for i in self.children],
            "args": self.args,
            "kwargs": self.kwargs,
            "tags": self.tags,
        }
        base.update(self._custom_dict())

        return base

    def accept(self, visitor, *args, **kwargs):
        # Some node types contain a dot to allow templates
        # to be created in a hierarchy of directories
        # but dots are not allowed in function names
        method_name = f"_visit_{self.node_type.replace('.', '__')}"

        try:
            method = getattr(visitor, method_name)
        except AttributeError:
            method = getattr(visitor, "_visit_default")

        return method(self, *args, **kwargs)

    def __eq__(self, other):
        try:
            return self.asdict() == other.asdict()
        except AttributeError:  # pragma: no cover
            return False

    def __repr__(self):  # pragma: no cover
        return str(self.asdict())

    def hash(self):  # pragma: no cover
        return hashlib.md5(str(self).encode("utf-8")).hexdigest()[:8]


class ValueNode(Node):
    node_type = "value_node"

    def __init__(
        self,
        value,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.value = value

    def _custom_dict(self):
        return {
            "value": self.value,
        }



import pytest
from mau.errors import MauErrorException
from mau.lexers.main_lexer import MainLexer
from mau.nodes.block import BlockNode
from mau.nodes.inline import TextNode
from mau.nodes.paragraph import ParagraphNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_parse_block_with_empty_body():
    source = """
    ----
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype=None,
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_parse_block_content():
    source = """
    ----
    This is a paragraph.
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype=None,
            children=[
                ParagraphNode(
                    children=[
                        TextNode("This is a paragraph."),
                    ]
                ),
            ],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_parse_block_content_variables():
    source = """
    ----
    :answer:42
    The answer is {answer}.
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype=None,
            children=[
                ParagraphNode(
                    children=[
                        TextNode("The answer is 42."),
                    ]
                ),
            ],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_parse_block_content_external_variables():
    source = """
    :answer:42
    ----
    The answer is {answer}.
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype=None,
            children=[
                ParagraphNode(
                    children=[
                        TextNode("The answer is 42."),
                    ]
                ),
            ],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]


def test_block_without_closing_fences():
    with pytest.raises(MauErrorException):
        runner("----")


def test_parse_block_secondary_content():
    source = """
    ----
    ----
    This is a paragraph that gets eaten.
    This is a second paragraph that gets eaten.

    This paragraph appears in the output.
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype=None,
            children=[],
            secondary_children=[
                ParagraphNode(
                    children=[
                        TextNode(
                            (
                                "This is a paragraph that gets eaten. "
                                "This is a second paragraph that gets eaten."
                            )
                        ),
                    ]
                ),
            ],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        ),
        ParagraphNode(
            children=[
                TextNode("This paragraph appears in the output."),
            ]
        ),
    ]


def test_parse_block_inside_block():
    source = """
    ----
    ++++
    ++++
    ----
    """

    assert runner(source).nodes == [
        BlockNode(
            subtype=None,
            children=[
                BlockNode(
                    subtype=None,
                    children=[],
                    secondary_children=[],
                    classes=[],
                    title=None,
                    engine=None,
                    preprocessor="none",
                    args=[],
                    kwargs={},
                )
            ],
            secondary_children=[],
            classes=[],
            title=None,
            engine=None,
            preprocessor="none",
            args=[],
            kwargs={},
        )
    ]



from mau.environment.environment import Environment
from mau.nodes.footnotes import FootnoteNode, FootnotesNode
from mau.nodes.inline import TextNode
from mau.visitors.base_visitor import BaseVisitor


def test_footnotes():
    visitor = BaseVisitor(Environment())

    node = FootnotesNode(
        children=[
            FootnoteNode(
                children=[TextNode("Footnote 1")],
                number="1",
                reference_anchor="anchor-1",
                content_anchor="anchor-1-def",
            ).to_entry(),
            FootnoteNode(
                children=[TextNode("Footnote 2")],
                number="2",
                reference_anchor="anchor-2",
                content_anchor="anchor-2-def",
            ).to_entry(),
        ],
        args=["arg1", "arg2"],
        kwargs={"key1": "value1"},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "footnotes",
            "entries": [
                {
                    "data": {
                        "content": [
                            {
                                "data": {
                                    "type": "text",
                                    "value": "Footnote 1",
                                    "args": [],
                                    "kwargs": {},
                                    "subtype": None,
                                    "tags": [],
                                }
                            },
                        ],
                        "content_anchor": "anchor-1-def",
                        "number": "1",
                        "reference_anchor": "anchor-1",
                        "type": "footnotes_entry",
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                },
                {
                    "data": {
                        "content": [
                            {
                                "data": {
                                    "type": "text",
                                    "value": "Footnote 2",
                                    "args": [],
                                    "kwargs": {},
                                    "subtype": None,
                                    "tags": [],
                                }
                            },
                        ],
                        "content_anchor": "anchor-2-def",
                        "number": "2",
                        "reference_anchor": "anchor-2",
                        "type": "footnotes_entry",
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                },
            ],
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1"},
            "tags": ["tag1", "tag2"],
            "subtype": None,
        }
    }



from mau.nodes.nodes import Node


class MacroNode(Node):
    """This node contains a macro, with a name and arguments."""

    node_type = "macro"

    def __init__(
        self,
        name,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.name = name

    def _custom_dict(self):
        return {
            "name": self.name,
        }


class MacroClassNode(Node):
    """Text with one or more classes."""

    node_type = "macro.class"

    def __init__(
        self,
        classes,
        parent=None,
        parent_position=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.classes = classes

    def _custom_dict(self):
        return {
            "classes": self.classes,
        }


class MacroLinkNode(Node):
    """This node contains a link."""

    node_type = "macro.link"

    def __init__(
        self,
        target,
        parent_position=None,
        parent=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.target = target

    def _custom_dict(self):
        return {
            "target": self.target,
        }


class MacroImageNode(Node):
    """This node contains an inline image."""

    node_type = "macro.image"

    def __init__(
        self,
        uri,
        alt_text=None,
        width=None,
        height=None,
        parent_position=None,
        parent=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.uri = uri
        self.alt_text = alt_text
        self.width = width
        self.height = height

    def _custom_dict(self):
        return {
            "uri": self.uri,
            "alt_text": self.alt_text,
            "width": self.width,
            "height": self.height,
        }


class MacroHeaderNode(Node):
    """This node contains a link to a header node."""

    node_type = "macro.header"

    def __init__(
        self,
        header_id,
        header=None,
        parent_position=None,
        parent=None,
        children=None,
        subtype=None,
        args=None,
        kwargs=None,
        tags=None,
    ):
        super().__init__(
            parent=parent,
            parent_position=parent_position,
            children=children,
            subtype=subtype,
            args=args,
            kwargs=kwargs,
            tags=tags,
        )
        self.header_id = header_id
        self.header = header

    def _custom_dict(self):
        return {
            "header_id": self.header_id,
            "header": self.header,
        }



from mau.environment.environment import Environment
from mau.nodes.page import HorizontalRuleNode
from mau.visitors.jinja_visitor import JinjaVisitor


def test_page_horizontal_rule_node():
    templates = {
        "text.j2": "{{ value }}",
        "horizontal_rule.j2": (
            "--- - {{ args | join(',') }} - "
            "{% for key, value in kwargs|items %}{{ key }}:{{ value }}{% endfor %} - "
            "{{ tags | join(',') }}"
        ),
    }

    environment = Environment()
    environment.update(templates, "mau.visitor.custom_templates")

    visitor = JinjaVisitor(environment)

    args = ["arg1", "arg2"]
    kwargs = {"key1": "value1"}
    tags = ["tag1", "tag2"]
    node = HorizontalRuleNode(args=args, kwargs=kwargs, tags=tags)

    result = visitor.visit(node)

    assert result == "--- - arg1,arg2 - key1:value1 - tag1,tag2"



from mau.lexers.main_lexer import MainLexer
from mau.nodes.page import HorizontalRuleNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_horizontal_rule():
    source = "---"

    expected = [HorizontalRuleNode()]

    assert runner(source).nodes == expected


def test_horizontal_rule_with_arguments():
    source = """
    [*break,arg1,key1=value1]
    ---
    """

    expected = [
        HorizontalRuleNode(
            subtype="break",
            args=["arg1"],
            kwargs={
                "key1": "value1",
            },
        ),
    ]

    assert runner(source).nodes == expected



from mau.nodes.block import BlockNode


def test_block_node():
    node = BlockNode(subtype="sometype")

    assert node.subtype == "sometype"
    assert node.children == []
    assert node.secondary_children == []
    assert node.args == []
    assert node.kwargs == {}
    assert node.node_type == "block"
    assert node == BlockNode(subtype="sometype")



from mau.lexers.base_lexer import TokenTypes
from mau.lexers.text_lexer import TextLexer
from mau.tokens.tokens import Token

from tests.helpers import init_lexer_factory, lexer_runner_factory

init_lexer = init_lexer_factory(TextLexer)

runner = lexer_runner_factory(TextLexer)


def test_normal_text():
    lex = runner("Normal text")

    assert lex.tokens == [
        Token(TokenTypes.TEXT, "Normal"),
        Token(TokenTypes.TEXT, " "),
        Token(TokenTypes.TEXT, "text"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_underscore():
    lex = runner("_underscore_")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "_"),
        Token(TokenTypes.TEXT, "underscore"),
        Token(TokenTypes.LITERAL, "_"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_star():
    lex = runner("*star*")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "*"),
        Token(TokenTypes.TEXT, "star"),
        Token(TokenTypes.LITERAL, "*"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_caret():
    lex = runner("^caret^")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "^"),
        Token(TokenTypes.TEXT, "caret"),
        Token(TokenTypes.LITERAL, "^"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_tilde():
    lex = runner("~tilde~")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "~"),
        Token(TokenTypes.TEXT, "tilde"),
        Token(TokenTypes.LITERAL, "~"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_backtick():
    lex = runner("`backtick`")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "`"),
        Token(TokenTypes.TEXT, "backtick"),
        Token(TokenTypes.LITERAL, "`"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_dollar():
    lex = runner("$dollar$")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "$"),
        Token(TokenTypes.TEXT, "dollar"),
        Token(TokenTypes.LITERAL, "$"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_percent():
    lex = runner("%percent%")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "%"),
        Token(TokenTypes.TEXT, "percent"),
        Token(TokenTypes.LITERAL, "%"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_curly_braces():
    lex = runner("{curly}")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "{"),
        Token(TokenTypes.TEXT, "curly"),
        Token(TokenTypes.LITERAL, "}"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_round_brackets():
    lex = runner("(round)")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "("),
        Token(TokenTypes.TEXT, "round"),
        Token(TokenTypes.LITERAL, ")"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_square_brackets():
    lex = runner("[square]")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "["),
        Token(TokenTypes.TEXT, "square"),
        Token(TokenTypes.LITERAL, "]"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_macro():
    lex = runner("[macro](value1,value2)")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "["),
        Token(TokenTypes.TEXT, "macro"),
        Token(TokenTypes.LITERAL, "]"),
        Token(TokenTypes.LITERAL, "("),
        Token(TokenTypes.TEXT, "value1,value2"),
        Token(TokenTypes.LITERAL, ")"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_macro_named_attributes():
    lex = runner("[macro](attr1=value1,attr2=value2)")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "["),
        Token(TokenTypes.TEXT, "macro"),
        Token(TokenTypes.LITERAL, "]"),
        Token(TokenTypes.LITERAL, "("),
        Token(TokenTypes.TEXT, "attr1=value1,attr2=value2"),
        Token(TokenTypes.LITERAL, ")"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_escaped_underscore():
    lex = runner(r"\_underscore\_")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "_"),
        Token(TokenTypes.TEXT, "underscore"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "_"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_escaped_square_brackets():
    lex = runner(r"\[square\]")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "["),
        Token(TokenTypes.TEXT, "square"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "]"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_escaped_round_brackets():
    lex = runner(r"\(round\)")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "("),
        Token(TokenTypes.TEXT, "round"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, ")"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_escaped_curly_braces():
    lex = runner(r"\{curly\}")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "{"),
        Token(TokenTypes.TEXT, "curly"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "}"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_escaped_quotes():
    lex = runner(r"\"quotes\"")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, '"'),
        Token(TokenTypes.TEXT, "quotes"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, '"'),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_escaped_backticks():
    lex = runner(r"\`backticks\`")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "`"),
        Token(TokenTypes.TEXT, "backticks"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "`"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_escaped_dollar():
    lex = runner(r"\$dollar\$")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "$"),
        Token(TokenTypes.TEXT, "dollar"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "$"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]


def test_escaped_percent():
    lex = runner(r"\%pecent\%")

    assert lex.tokens == [
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "%"),
        Token(TokenTypes.TEXT, "pecent"),
        Token(TokenTypes.LITERAL, "\\"),
        Token(TokenTypes.LITERAL, "%"),
        Token(TokenTypes.EOL),
        Token(TokenTypes.EOF),
    ]



import pytest
from mau.environment.environment import Environment
from mau.errors import MauErrorException
from mau.lexers.main_lexer import MainLexer
from mau.nodes.inline import StyleNode, TextNode
from mau.nodes.page import ContainerNode, DocumentNode
from mau.nodes.paragraph import ParagraphNode
from mau.nodes.toc import TocNode
from mau.parsers.main_parser import MainParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


def test_parse_discards_empty_lines():
    source = "\n\n\n\n"

    assert runner(source).nodes == []


def test_parse_output():
    source = ""

    assert runner(source).output == {
        "content": ContainerNode(children=[]),
        "toc": ContainerNode(children=[TocNode()]),
    }


def test_parse_output_custom_content_container():
    source = ""

    environment = Environment()
    document = DocumentNode()
    environment.setvar("mau.parser.content_wrapper", document)

    assert runner(source, environment).output == {
        "content": document,
        "toc": ContainerNode(children=[TocNode()]),
    }


def test_parse_output_custom_toc_container():
    source = ""

    environment = Environment()
    document = DocumentNode(children=[TocNode()])
    environment.setvar("mau.parser.toc_wrapper", document)

    assert runner(source, environment).output == {
        "content": ContainerNode(children=[]),
        "toc": document,
    }


def test_parse_single_line_comments():
    source = "// Just a comment"

    assert runner(source).nodes == []


def test_parse_multi_line_comments():
    source = """
    ////
    This is a
    multiline
    comment
    ////
    """

    assert runner(source).nodes == []


def test_parse_multi_line_comments_with_dangerous_characters():
    source = """
    ////
    .This is a
    [multiline]
    ----
    comment
    ////
    """

    assert runner(source).nodes == []


def test_parse_open_multi_line_comments():
    source = """
    ////
    .This is a
    [multiline]
    ----
    comment
    """

    with pytest.raises(MauErrorException):
        runner(source)


def test_command():
    source = "::somecommand:arg1, arg2, name1=value1, name2=value2"

    assert runner(source).nodes == []


def test_command_without_arguments():
    source = "::somecommand:"

    assert runner(source).nodes == []


def test_style_underscore():
    source = """
    This is _underscore_ text
    """

    assert runner(source).nodes == [
        ParagraphNode(
            children=[
                TextNode("This is "),
                StyleNode(
                    value="underscore",
                    children=[
                        TextNode("underscore"),
                    ],
                ),
                TextNode(" text"),
            ],
        )
    ]


def test_style_at_beginning():
    source = """
    *This is star text*
    """

    assert runner(source).nodes == [
        ParagraphNode(
            children=[
                StyleNode(
                    value="star",
                    children=[
                        TextNode("This is star text"),
                    ],
                ),
            ],
        )
    ]


def test_style_not_closed():
    source = r"""
    This ` is a backtick and this _an underscore
    """

    assert runner(source).nodes == [
        ParagraphNode(
            children=[
                TextNode("This ` is a backtick and this _an underscore"),
            ],
        )
    ]


def test_style_escape_markers():
    source = r"""
    This is \_underscore\_ and this is \`verbatim\`
    """

    assert runner(source).nodes == [
        ParagraphNode(
            children=[
                TextNode("This is _underscore_ and this is `verbatim`"),
            ],
        )
    ]



from mau.lexers.text_lexer import TextLexer
from mau.nodes.footnotes import FootnoteNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_macro_footnote():
    source = "[footnote](notename)"

    footnote_node = FootnoteNode()
    expected = [footnote_node]

    parser = runner(source)
    assert parser.nodes == expected
    assert parser.footnotes == {"notename": footnote_node}



from unittest.mock import patch

import pytest
from mau.environment.environment import Environment
from mau.errors import MauErrorException
from mau.lexers.main_lexer import MainLexer
from mau.nodes.header import HeaderNode
from mau.nodes.inline import SentenceNode, TextNode
from mau.parsers.main_parser import MainParser, header_anchor

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(MainLexer, MainParser)

runner = parser_runner_factory(MainLexer, MainParser)


@patch("mau.parsers.toc.hashlib.md5")
def test_default_header_anchor_function(mock_md5):
    mock_md5().hexdigest.return_value = "XXYY"

    assert header_anchor("Some Words 1234 56", "1") == "some-words-1234-56-XXYY"


@patch("mau.parsers.toc.hashlib.md5")
def test_default_header_anchor_function_multiple_spaces(mock_md5):
    mock_md5().hexdigest.return_value = "XXYY"

    assert (
        header_anchor("Some    Words     1234    56", "1") == "some-words-1234-56-XXYY"
    )


@patch("mau.parsers.toc.hashlib.md5")
def test_default_header_anchor_function_filter_characters(mock_md5):
    mock_md5().hexdigest.return_value = "XXYY"

    assert header_anchor("Some #Words @ 12!34 56", "1") == "some-words-1234-56-XXYY"


def test_custom_header_anchor_function():
    source = """
    = Title of the section
    """

    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: "XXXXXY"
    )

    assert runner(source, environment).nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Title of the section")]),
            level="1",
            anchor="XXXXXY",
        )
    ]


def test_parse_header_level_1():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: "XXXXXY"
    )

    source = """
    = Title of the section
    """

    assert runner(source, environment).nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Title of the section")]),
            level="1",
            anchor="XXXXXY",
        )
    ]


def test_parse_header_level_3():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: "XXXXXX"
    )

    source = """
    === Title of a subsection
    """

    assert runner(source, environment).nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Title of a subsection")]),
            level="3",
            anchor="XXXXXX",
        )
    ]


def test_parse_collect_headers():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: f"{text}-XXXXXX"
    )

    source = """
    = Header 1
    == Header 1.1
    == Header 1.2
    = Header 2
    == Header 2.1
    === Header 2.1.1
    """

    parser = runner(source, environment)

    assert parser.nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1")]),
            level="1",
            anchor="Header 1-XXXXXX",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1.1")]),
            level="2",
            anchor="Header 1.1-XXXXXX",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1.2")]),
            level="2",
            anchor="Header 1.2-XXXXXX",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2")]),
            level="1",
            anchor="Header 2-XXXXXX",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2.1")]),
            level="2",
            anchor="Header 2.1-XXXXXX",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2.1.1")]),
            level="3",
            anchor="Header 2.1.1-XXXXXX",
        ),
    ]

    assert parser.toc_manager.headers == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1")]),
            level="1",
            anchor="Header 1-XXXXXX",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1.1")]),
            level="2",
            anchor="Header 1.1-XXXXXX",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 1.2")]),
            level="2",
            anchor="Header 1.2-XXXXXX",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2")]),
            level="1",
            anchor="Header 2-XXXXXX",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2.1")]),
            level="2",
            anchor="Header 2.1-XXXXXX",
        ),
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header 2.1.1")]),
            level="3",
            anchor="Header 2.1.1-XXXXXX",
        ),
    ]


def test_attributes_header():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: f"{text}-XXXXXX"
    )

    source = """
    [arg1,key1=value1]
    = Header
    """

    assert runner(source, environment).nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header")]),
            level="1",
            anchor="Header-XXXXXX",
            args=["arg1"],
            kwargs={"key1": "value1"},
        ),
    ]


def test_header_attributes_can_overwrite_anchor():
    source = """
    [arg1, anchor=someheader, key1=value1]
    = Header
    """

    assert runner(source).nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header")]),
            level="1",
            anchor="someheader",
            args=["arg1"],
            kwargs={"key1": "value1"},
        ),
    ]


def test_header_with_id_is_stored():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: f"{text}-XXXXXX"
    )

    source = """
    [arg1,id=someid,key1=value1]
    = Header
    """

    node = HeaderNode(
        value=SentenceNode(children=[TextNode("Header")]),
        level="1",
        anchor="Header-XXXXXX",
        args=["arg1"],
        kwargs={"id": "someid", "key1": "value1"},
    )

    parser = runner(source, environment)

    assert parser.nodes == [node]
    assert parser.internal_links_manager.headers == {"someid": node}


def test_header_with_duplicate_id():
    source = """
    [arg1,id=someid,key1=value1]
    = Header

    [id=someid]
    = Another Header
    """

    with pytest.raises(MauErrorException):
        runner(source)


def test_single_tag_header():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: f"{text}-XXXXXX"
    )

    source = """
    [arg1, #tag1, key1=value1]
    = Header
    """

    assert runner(source, environment).nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header")]),
            level="1",
            anchor="Header-XXXXXX",
            args=["arg1"],
            kwargs={"key1": "value1"},
            tags=["tag1"],
        ),
    ]


def test_headeer_subtype():
    environment = Environment()
    environment.setvar(
        "mau.parser.header_anchor_function", lambda text, level: f"{text}-XXXXXX"
    )

    source = """
    [*type1]
    = Header
    """

    parser = runner(source, environment)
    assert parser.nodes == [
        HeaderNode(
            value=SentenceNode(children=[TextNode("Header")]),
            level="1",
            anchor="Header-XXXXXX",
            args=[],
            kwargs={},
            tags=[],
            subtype="type1",
        ),
    ]

    header_node = parser.nodes[0]
    text_node = header_node.value.children[0]

    assert text_node.parent == header_node



from mau.lexers.text_lexer import TextLexer
from mau.nodes.inline import TextNode
from mau.nodes.macros import MacroNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_macro():
    source = "[macroname](value1,value2)"

    expected = [
        MacroNode(
            "macroname",
            args=["value1", "value2"],
        ),
    ]

    assert runner(source).nodes == expected


def test_incomplete_macro():
    source = "[macroname](value1"

    expected = [
        TextNode(
            "[macroname](value1",
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_arguments_with_quotes():
    source = '[macroname]("value1,value2")'

    expected = [
        MacroNode(
            "macroname",
            args=["value1,value2"],
        ),
    ]

    assert runner(source).nodes == expected


def test_macro_named_arguments():
    source = "[macroname](name,arg1=value1)"

    expected = [
        MacroNode(
            "macroname",
            args=["name"],
            kwargs={"arg1": "value1"},
        ),
    ]

    assert runner(source).nodes == expected



from mau.environment.environment import Environment
from mau.nodes.inline import TextNode
from mau.nodes.lists import ListItemNode, ListNode
from mau.visitors.base_visitor import BaseVisitor


def test_list_item_node():
    visitor = BaseVisitor(Environment())

    node = ListItemNode(level="4", children=[TextNode("Just some text.")])

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "list_item",
            "level": 4,
            "content": [
                {
                    "data": {
                        "type": "text",
                        "value": "Just some text.",
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                }
            ],
            "args": [],
            "kwargs": {},
            "subtype": None,
            "tags": [],
        }
    }


def test_list_node():
    visitor = BaseVisitor(Environment())

    node = ListNode(
        ordered=True,
        main_node=True,
        start=42,
        children=[ListItemNode(level="4", children=[TextNode("Just some text.")])],
        args=["arg1", "arg2"],
        kwargs={"key1": "value1", "start": 4},
        tags=["tag1", "tag2"],
    )

    result = visitor.visit(node)

    assert result == {
        "data": {
            "type": "list",
            "start": 42,
            "items": [
                {
                    "data": {
                        "type": "list_item",
                        "level": 4,
                        "content": [
                            {
                                "data": {
                                    "type": "text",
                                    "value": "Just some text.",
                                    "args": [],
                                    "kwargs": {},
                                    "subtype": None,
                                    "tags": [],
                                }
                            }
                        ],
                        "args": [],
                        "kwargs": {},
                        "subtype": None,
                        "tags": [],
                    }
                }
            ],
            "main_node": True,
            "ordered": True,
            "args": ["arg1", "arg2"],
            "kwargs": {"key1": "value1", "start": 4},
            "tags": ["tag1", "tag2"],
            "subtype": None,
        }
    }



from mau.nodes.nodes import Node, ValueNode


class WordNode(ValueNode):
    """This is a single word, it's used internally
    and eventually packed together with others into
    a TextNode
    """

    node_type = "word"


class TextNode(ValueNode):
    """This contains plain text and is created
    as a collation of multiple WordNode objects
    """

    node_type = "text"


class RawNode(ValueNode):
    """This contains plain text but the content
    should be treated as raw data and left untouched.
    E.g. it shouldn't be escaped.
    """

    node_type = "raw"


class VerbatimNode(ValueNode):
    """This node contains verbatim text."""

    node_type = "verbatim"


class SentenceNode(Node):
    """A recursive container node.

    This node represents the content of a paragraph, but it is recursive,
    while ParagraphNode is not.
    """

    node_type = "sentence"


class StyleNode(ValueNode):
    """Describes the style applied to a node."""

    node_type = "style"



from mau.lexers.text_lexer import TextLexer
from mau.nodes.inline import StyleNode, TextNode, VerbatimNode
from mau.nodes.macros import MacroClassNode
from mau.parsers.text_parser import TextParser

from tests.helpers import init_parser_factory, parser_runner_factory

init_parser = init_parser_factory(TextLexer, TextParser)

runner = parser_runner_factory(TextLexer, TextParser)


def test_single_class():
    source = 'Some text [class]("text with that class", classname)'

    expected = [
        TextNode("Some text "),
        MacroClassNode(
            classes=["classname"],
            children=[
                TextNode("text with that class"),
            ],
        ),
    ]

    assert runner(source).nodes == expected


def test_multiple_classes():
    source = 'Some text [class]("text with that class", classname1, classname2)'

    expected = [
        TextNode("Some text "),
        MacroClassNode(
            classes=["classname1", "classname2"],
            children=[
                TextNode("text with that class"),
            ],
        ),
    ]

    assert runner(source).nodes == expected


def test_parse_class_with_rich_text():
    source = '[class]("Some text with `verbatim words` and _styled ones_", classname)'

    expected = [
        MacroClassNode(
            classes=["classname"],
            children=[
                TextNode("Some text with "),
                VerbatimNode("verbatim words"),
                TextNode(" and "),
                StyleNode(value="underscore", children=[TextNode("styled ones")]),
            ],
        ),
    ]

    assert runner(source).nodes == expected
